function BN_FL_2Neuron_threshold(seed)

% Simulates the FL algorithm in a 2 Bayesian Spiking Neuron (BSN) 
% network with N inputs going to the 1st neuron and the 1st neuron providing the only
% input to the 2nd neuron. The main aim of this code is to use 'seed' to
% specify both random true parameter values as well as specific dynamic
% threshold parameter values (i.e. values for theta_U and theta_D in the paper)

% This function is similar in structure to 'BN_FL_2Neuron.m' the main 
% difference being setting 'run_type = 4'. There are other minor differences which do not change the way the code runs.
% An example of a minor difference is that you can't track the convergence of the
% parameters in this code which makes this code a little more efficient.
% Please refer to the comments used in 'BN_FL_2Neuron.m' to understand the
% variables and code segments. 

% inputs: 
% 'seed': seeds the randomisation of the true parameter values for the
% transition and observation rates. The seed also chooses one of the 25 (theta_U,theta_D) pairs that define the dynamic threshold 
% and controlled by 'th_seed' and 'th_map' where 'th_seed' indexes
% 'th_map'. This enabled us to do 100 different simulations for each of the
% 25 (theta_U,theta_D) pairs in the paper, one 'seed' at a time using this function.

% outputs: There are no outputs, however, there are options to create simulation 
% figures by setting 'show_figs' to 1, and simulated variables are automatically saved to a .mat file 

% In this function the parameter estimates are initially set to multiplicatively scaled values of the true
% parameters. This is controlled by the selection of the 'run_type' variable.

% Note rates are given in units of spikes/(0.1 ms). Therefore be aware of the following
% conversion: r = 1 spikes/s = 0.001 spikes/ms = 0.001 spikes/ 10 x 0.1 ms = 0.0001 spikes/(0.1 ms)

% This code includes additional code/checks/options that can be removed for
% efficiency purposes but have been left here. The preference has been to
% just give adequate comments. Follow the comments in a sequential order to
% make the most sense of the code. It is recommended to make your own simplified code
% but to use this as a guide.

% For help email: levink@unimelb.edu.au

% Copyright 15/11/2013 Levin Kuhlmann

show_figs = 0;

tic;

run_type = 4; %1 or 2 or 3 or 4

if (run_type == 1) % single set of true parameters, run each job for 1,000,000 time steps and 10 different perturbations and 100 trials each perturbation
save_path = 'on_avs_data_1mil/'
rand_seed = mod(seed-1,100) + 1 % 100 random seeds
perturb = mod(ceil(seed/100)-1,10) + 1 % 10 perturbation windows
rate_select = mod(ceil(seed/1000)-1,2) + 1 % 2 sample number factors
sample_select = ceil(seed/2000) % 2 rate factors
T_steps = 1000000;
true_init = 0;
elseif (run_type == 2) %  single set of true parameters, run each job for 10,000,000 time steps and 10 different perturbations and 100 trials each perturbation
save_path = 'on_avs_data_10mil/'
rand_seed = seed % 1000 random seeds
perturb = 10 % 10 perturbation windows
rate_select = 1 % 2 sample number factors
sample_select = 1 % 2 rate factors
T_steps = 10000000;
true_init = 1;
elseif (run_type == 3) % run each job for 1,000,000 time steps and 1000 different true parameter values between 0 to 0.1 and 
% using initial parameter perturbations within x to y percent
save_path = 'on_avs_data_rand/'
rand_seed = seed % 1000 random seeds
perturb = 10 % 10 perturbation windows
rate_select = 1 % 2 sample number factors
sample_select = 1 % 2 rate factors
T_steps = 1000000;
true_init = 1;
elseif (run_type == 4) %  single set of true parameters, run each job for 1,000,000 time steps for upper threshold 0.5 to 0.9 in steps of 0.1
% and for lower thresholds 0.1 to 0.5 in steps of 0.1, making 5 x 5 =
% 25 combinations running 40 trials for each combination and using initial
% parameter perturbations within x to y percent
save_path = 'on_avs_data_th/'
rand_seed = mod(seed-1,100) + 1 % 100 random seeds
th_seed = mod(ceil(seed/100)-1,25) + 1 % 25 th seeds
perturb = 2 % 10 perturbation windows
rate_select = 1 % 2 sample number factors
sample_select = 1 % 2 rate factors
T_steps = 1000000;
true_init = 1;%0;
th_map = [ 1 2 3 4 5 1 2 3 4 5 1 2 3 4 5 1 2 3 4 5 1 2 3 4 5 ; 1 1 1 1 1 2 2 2 2 2 3 3 3 3 3 4 4 4 4 4 5 5 5 5 5 ];
elseif (run_type == 5) % use new form of perturbation with a minimum. single set of true parameters, run each job for 1,000,000 time steps and 10 different perturbations and 100 trials each perturbation
save_path = 'on_avs_data_1mil_perturb2/' 
rand_seed = mod(seed-1,100) + 1 % 100 random seeds
perturb = mod(ceil(seed/100)-1,10) + 1 % 10 perturbation windows
rate_select = mod(ceil(seed/1000)-1,2) + 1 % 2 sample number factors
sample_select = ceil(seed/2000) % 2 rate factors
T_steps = 1000000;
true_init = 0;
end

    
% initialise
N_levels = 3; % number of neuron layers/levels. includes input layer
dt = .1; % ms
if (sample_select == 1)
    eta = 1/100000;%10^(-5);
elseif (sample_select == 2)
    eta = 10^(-6);
end 

T_stat = 1*10^5;
rand('state',rand_seed)
N_track = 100000;
step_disp = 100000;
use_crit_LL = 1;
if (use_crit_LL == 1)
    stop_criterion = 0.00005;
elseif (use_crit_LL == 0)
    stop_criterion = 0;
end
crit_perc = 0.01;
perc_low = [ 0 0.2 0.4 0.6 0.8 1.0 1.2 1.4 1.6 1.8 ];
perc_upp = [ 0.2 0.4 0.6 0.8 1.0 1.2 1.4 1.6 1.8 2.0 ];
perc_sc = [ 3 5 ];
p_up_array = [ 0.55 0.65 0.75 0.85 0.95 ];
p_down_array = [ 0.05 0.15 0.25 0.35 0.45 ];
rate_factors = [ 1 10 ];
sample_factors = [ 1 10 ];
step_crit = 100000;

if (run_type == 4)
    p_up_sc = p_up_array(th_map(1,th_seed))
    p_down_sc = p_down_array(th_map(2,th_seed))
elseif (run_type == 3)
    p_up_sc = 0.75
    p_down_sc = 0.25
else
    p_up_sc = 0.75
    p_down_sc = 0.25
end

T = T_steps*sample_factors(sample_select)%5*10^5; % number of time steps
g_o = 1.45;
level(3).N_neurons = 1;
level(2).N_neurons = 1;
level(1).N_inputs = 20;
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        if (l == 2)
            N_tmp = level(1).N_inputs;
        else
            N_tmp = level(l).N_neurons;
        end
        level(l).neuron(n).connect = ones(1,N_tmp); % determines connections to previous layer
        level(l).neuron(n).ind_connect = find(level(l).neuron(n).connect == 1);
        level(l).neuron(n).N_connects = sum(level(l).neuron(n).connect); % number of input synapses
        level(l).neuron(n).q = [ 0.5 ; 0.5 ]; % q corresponds to pi the initial conditions for the state probabilities
    end
end
% r*dt and lambda*dt need to be less than 1, make sure rates are large
% enough or T is large enough to get enough statistics
if ((run_type < 2) & (run_type > 4))
    level(3).neuron(1).r_on = 0.001*rate_factors(rate_select);%500;
    level(3).neuron(1).r_off = 0.01*rate_factors(rate_select);%1000;
    level(2).neuron(1).r_on = 0.001*rate_factors(rate_select);%500;
    level(2).neuron(1).r_off = 0.01*rate_factors(rate_select);%1000;
elseif ((run_type >= 2) & (run_type <= 4))
    level(2).neuron(1).r_off = (0.001+rand(1)*9.999)*rate_factors(rate_select);%1000;%0.114
    level(2).neuron(1).r_on = (0.001+rand(1)*9.999)*rate_factors(rate_select);
    level(3).neuron(1).r_on = level(2).neuron(1).r_on;%500;
    level(3).neuron(1).r_off = level(2).neuron(1).r_off;%1000;
    true_r_on = level(2).neuron(1).r_on
    true_r_off = level(2).neuron(1).r_off
end
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        level(l).neuron(n).a = [ (1-level(l).neuron(n).r_on*dt) level(l).neuron(n).r_on*dt ; level(l).neuron(n).r_off*dt (1-level(l).neuron(n).r_off*dt) ];
    end
end
for s = 1:level(1).N_inputs
    if ((run_type < 2) & (run_type > 4))
        level(1).input(s).lambda_on = 0.05*rate_factors(rate_select);%.08*1; 
        level(1).input(s).lambda_off = 0.01*rate_factors(rate_select);
    elseif ((run_type >= 2) & (run_type <= 4))
        level(1).input(s).lambda_on = (0.001+rand(1)*9.999)*rate_factors(rate_select);%.08*1;%0.999
        level(1).input(s).lambda_off = (0.001+rand(1)*9.999)*rate_factors(rate_select);
        true_q_on = level(1).input(s).lambda_on
        true_q_off = level(1).input(s).lambda_off
    end
    level(1).input(s).b = [ (1-level(1).input(s).lambda_off*dt) level(1).input(s).lambda_off*dt ; (1-level(1).input(s).lambda_on*dt) level(1).input(s).lambda_on*dt ];
end

% create EM variables for each neuron
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        level(l).neuron(n).r_hat_trace = NaN(2,N_track);
        for s = 1:level(l).neuron(n).N_connects
            level(l).neuron(n).syn(s).q_hat_trace = NaN(2,N_track);
        end
        level(l).neuron(n).r_on_hat = [];
        level(l).neuron(n).r_off_hat = [];
    end
end

% initialise 
if (true_init == 0) % initialise with random values
    for l = 2:N_levels
        for n = 1:level(l).N_neurons
            rate_min = 1*10^(-6);
            ron_percl = perc_low(perturb)*level(l).neuron(n).r_on;
            ron_percu = perc_upp(perturb)*level(l).neuron(n).r_on;
            ron_t1 = level(l).neuron(n).r_on + ron_percl;
            ron_t2 = level(l).neuron(n).r_on + ron_percu;
            ron_b2 = max(level(l).neuron(n).r_on - ron_percl,rate_min);
            ron_b1 = max(level(l).neuron(n).r_on - ron_percu,rate_min);
            ratio_tb = (ron_t2-ron_t1)/(ron_t2-ron_t1+ron_b2-ron_b1);
            rndc = rand(1,1);
            if (rndc <= ratio_tb)
                rontr = ron_t1 + (ron_t2 - ron_t1)*rndc/ratio_tb
            else
                rontr = ron_b1 + (ron_b2 - ron_b1)*(1-rndc)/(1-ratio_tb)                
            end
            roff_percl = perc_low(perturb)*level(l).neuron(n).r_off;
            roff_percu = perc_upp(perturb)*level(l).neuron(n).r_off;
            roff_t1 = level(l).neuron(n).r_off + roff_percl;
            roff_t2 = level(l).neuron(n).r_off + roff_percu;
            roff_b2 = max(level(l).neuron(n).r_off - roff_percl,rate_min);
            roff_b1 = max(level(l).neuron(n).r_off - roff_percu,rate_min);
            ratio_tb = (roff_t2-roff_t1)/(roff_t2-roff_t1+roff_b2-roff_b1);
            rndc = rand(1,1);
            if (rndc <= ratio_tb)
                rofftr = roff_t1 + (roff_t2 - roff_t1)*rndc/ratio_tb
            else
                rofftr = roff_b1 + (roff_b2 - roff_b1)*(1-rndc)/(1-ratio_tb)                
            end
            
            if (run_type == 2) | (run_type == 3) | (run_type == 4)
                rontr = perc_sc(perturb)*level(l).neuron(n).r_on;
                rofftr = perc_sc(perturb)*level(l).neuron(n).r_off;
            end
            
            level(l).neuron(n).r_hat_trace(:,1) = [ rontr ; rofftr ];
            level(l).neuron(n).r_on_init = rontr;
            level(l).neuron(n).r_off_init = rofftr;           
            
            if (l == 2)
                qon_rate_min = 1*10^(-6);
                qoff_rate_min = 1*10^(-6);
            elseif (l == 3)
                qon_rate_min = 0.03;
                qoff_rate_min = 1*10^(-6);
            end
            
            for s = 1:level(l).neuron(n).N_connects
                if (l == 2)
                    lon_percl = perc_low(perturb)*level(1).input(s).lambda_on;
                    lon_percu = perc_upp(perturb)*level(1).input(s).lambda_on;
                    lon_t1 = level(1).input(s).lambda_on + lon_percl;
                    lon_t2 = level(1).input(s).lambda_on + lon_percu;
                    lon_b2 = max(level(1).input(s).lambda_on - lon_percl,qon_rate_min);
                    lon_b1 = max(level(1).input(s).lambda_on - lon_percu,qon_rate_min);
                else
                    lon_percl = perc_low(perturb)*0.3879;
                    lon_percu = perc_upp(perturb)*0.3879;
                    lon_t1 = 0.3879 + lon_percl;
                    lon_t2 = 0.3879 + lon_percu;
                    lon_b2 = max(0.3879 - lon_percl,qon_rate_min);
                    lon_b1 = max(0.3879 - lon_percu,qon_rate_min);
                end
                ratio_tb = (lon_t2-lon_t1)/(lon_t2-lon_t1+lon_b2-lon_b1);
                rndc = rand(1,1);
                if (rndc <= ratio_tb)
                    lon_rnd = lon_t1 + (lon_t2 - lon_t1)*rndc/ratio_tb
                else
                    lon_rnd = lon_b1 + (lon_b2 - lon_b1)*(1-rndc)/(1-ratio_tb)                
                end
                if (l == 2)
                    loff_percl = perc_low(perturb)*level(1).input(s).lambda_off;
                    loff_percu = perc_upp(perturb)*level(1).input(s).lambda_off;
                    loff_t1 = level(1).input(s).lambda_off + loff_percl;
                    loff_t2 = level(1).input(s).lambda_off + loff_percu;
                    loff_b2 = max(level(1).input(s).lambda_off - loff_percl,qoff_rate_min);
                    loff_b1 = max(level(1).input(s).lambda_off - loff_percu,qoff_rate_min);
                else
                    loff_percl = perc_low(perturb)*7.5285*10^-3;
                    loff_percu = perc_upp(perturb)*7.5285*10^-3;
                    loff_t1 = 7.5285*10^-3 + loff_percl;
                    loff_t2 = 7.5285*10^-3 + loff_percu;
                    loff_b2 = max(7.5285*10^-3 - loff_percl,qoff_rate_min);
                    loff_b1 = max(7.5285*10^-3 - loff_percu,qoff_rate_min);
                end                    
                ratio_tb = (loff_t2-loff_t1)/(loff_t2-loff_t1+loff_b2-loff_b1);
                rndc = rand(1,1);
                if (rndc <= ratio_tb)
                    loff_rnd = loff_t1 + (loff_t2 - loff_t1)*rndc/ratio_tb
                else
                    loff_rnd = loff_b1 + (loff_b2 - loff_b1)*(1-rndc)/(1-ratio_tb)                
                end
                
                if (run_type == 2) | (run_type == 3) | (run_type == 4)
                    if (l == 2)
                        lon_rnd = perc_sc(perturb)*level(1).input(s).lambda_on;
                        loff_rnd = perc_sc(perturb)*level(1).input(s).lambda_off;
                    else
                        lon_rnd = perc_sc(perturb)*level(1).input(s).lambda_on;
                        loff_rnd = perc_sc(perturb)*level(1).input(s).lambda_off;
                    end
                end
                
                level(l).neuron(n).syn(s).q_hat_trace(:,1) = [ lon_rnd ; loff_rnd ];
                level(l).neuron(n).syn(s).q_i_on_init = lon_rnd;
                level(l).neuron(n).syn(s).q_i_off_init = loff_rnd;
            end
        end
    end
elseif (true_init == 1) % initialise with true parameter values
    for l = 2:N_levels
        for n = 1:level(l).N_neurons
            level(l).neuron(n).r_hat_trace(:,1) = [ level(l).neuron(n).r_on ; level(l).neuron(n).r_off ];
            level(l).neuron(n).r_on_init = level(l).neuron(n).r_on;
            level(l).neuron(n).r_off_init = level(l).neuron(n).r_off;
            if (l == 2)
                for s = 1:level(l).neuron(n).N_connects
                    level(l).neuron(n).syn(s).q_hat_trace(:,1) = [ level(1).input(level(l).neuron(n).ind_connect(s)).lambda_on ; level(1).input(level(l).neuron(n).ind_connect(s)).lambda_off ];
                    level(l).neuron(n).syn(s).q_i_on_init = level(1).input(level(l).neuron(n).ind_connect(s)).lambda_on;
                    level(l).neuron(n).syn(s).q_i_off_init = level(1).input(level(l).neuron(n).ind_connect(s)).lambda_off;
                end
            else
                for s = 1:level(l).neuron(n).N_connects
                    level(l).neuron(n).syn(s).q_hat_trace(:,1) = [ level(1).input(level(l).neuron(n).ind_connect(s)).lambda_on ; level(1).input(level(l).neuron(n).ind_connect(s)).lambda_off ]; % these are undefined for the generative model so choose random values
                    level(l).neuron(n).syn(s).q_i_on_init = level(1).input(level(l).neuron(n).ind_connect(s)).lambda_on;
                    level(l).neuron(n).syn(s).q_i_off_init = level(1).input(level(l).neuron(n).ind_connect(s)).lambda_off;
                end
            end
        end
    end
end

% create plot data of HMM values
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        level(l).neuron(n).HMM_plot_r_on = repmat(level(l).neuron(n).r_on,[ 1 N_track ]);
        level(l).neuron(n).HMM_plot_r_off = repmat(level(l).neuron(n).r_off,[ 1 N_track ]);
        if (l == 2)
            for s = 1:level(l).neuron(n).N_connects
                level(l).neuron(n).syn(s).HMM_plot_q_on = repmat(level(1).input(level(l).neuron(n).ind_connect(s)).lambda_on,[ 1 N_track ]);
                level(l).neuron(n).syn(s).HMM_plot_q_off = repmat(level(1).input(level(l).neuron(n).ind_connect(s)).lambda_off,[ 1 N_track ]);
            end
        end
    end
end

% create arrays to store HMM simulation results and bayesian neuron outputs
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        level(l).neuron(n).x_k_trace = zeros(1,N_track); % neuron hidden state
        level(l).neuron(n).x_k_est_trace = zeros(1,N_track); % neuron hidden state
        if (l == 2)
            level(l).neuron(n).y_k_trace = zeros(1,T); % neuron output spikes
        else
            level(l).neuron(n).y_k_trace = zeros(1,N_track); % neuron output spikes
        end
    end        
end
for s = 1:level(1).N_inputs
    level(1).input(s).y_k_trace = zeros(1,N_track); % network input spikes/HMM observations
end

% run the FL algorithm for each level > 1 and each neuron in
% the corresponding level, run lower levels first
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        t_1 = toc;
        % use parameter estimates and numerically integrate L and G to determine output
        % spikes and run FL.
        % initialise L and G and define relevant parameters
        level(l).neuron(n).r_on_hat = level(l).neuron(n).r_hat_trace(1,1);
        level(l).neuron(n).r_off_hat = level(l).neuron(n).r_hat_trace(2,1);
        level(l).neuron(n).th = 0;
        for s = 1:level(l).neuron(n).N_connects
            level(l).neuron(n).syn(s).q_i_on_hat = level(l).neuron(n).syn(s).q_hat_trace(1,1);
            level(l).neuron(n).syn(s).q_i_off_hat = level(l).neuron(n).syn(s).q_hat_trace(2,1);
            level(l).neuron(n).syn(s).weight = log(level(l).neuron(n).syn(s).q_i_on_hat/level(l).neuron(n).syn(s).q_i_off_hat); 
            level(l).neuron(n).th = level(l).neuron(n).th + level(l).neuron(n).syn(s).q_i_on_hat - level(l).neuron(n).syn(s).q_i_off_hat;
        end
        ronhat = level(l).neuron(n).r_on_hat;
        roffhat = level(l).neuron(n).r_off_hat;
        rontrue = level(l).neuron(n).r_on;
        rofftrue = level(l).neuron(n).r_off;
        th = level(l).neuron(n).th;
        disp('initial values for:')
        disp([ 'level ' num2str(l) ', neuron ' num2str(n) ', ronhat ' num2str(ronhat) ', roffhat ' num2str(roffhat) ', th ' num2str(th) ])
        for s = 1:level(l).neuron(n).N_connects
            wght = level(l).neuron(n).syn(s).weight;
            disp([ 'synapse ' num2str(s) ', wght ' num2str(wght) ])
            syn(s).qion = level(l).neuron(n).syn(s).q_i_on_hat;
            syn(s).qioff = level(l).neuron(n).syn(s).q_i_off_hat;
        end
        
        L = zeros(1,N_track);
        L(1) = log(level(l).neuron(n).r_on_hat/level(l).neuron(n).r_off_hat);
        L_sm = L;
        L_up = L;
        L_lo = L;
        p_up = p_up_sc*ones(1,N_track);
        p_down = p_down_sc*ones(1,N_track);
        eta_sm = 1/5;
        G = L;
        if (l == 2)
            L_true = zeros(1,N_track);
            L_true(1) = L(1);
            L_true_flip = zeros(1,N_track);
            L_true_flip(1) = L(1);
        end
        % initialise additional EM terms
        t_on = 0.5;
        N_on_off = 0;
        N_off_on = 0;
        for s = 1:level(l).neuron(n).N_connects
            syn(s).N_on = 0;
            syn(s).N = 0;
        end

        if (show_figs == 1)
            figure('position',[ 30 50 900 600 ])
        end
        % run bayesian neuron and EM online
        level(l).neuron(n).crit_reached = 0;
        level(l).neuron(n).tot_EM_steps = NaN;
        first = 1;
        crit_p = 2 + 2*level(l).neuron(n).N_connects;
        rontrue = level(l).neuron(n).r_on;
        rofftrue = level(l).neuron(n).r_off;
        if (l == 2)
            for s = 1:level(l).neuron(n).N_connects
                syn(s).qion_true = level(1).input(s).lambda_on;
                syn(s).qioff_true = level(1).input(s).lambda_off;
            end
        end
        x_c = 0;
        N_off_on_c = 0;
        N_on_off_c = 0;
        p_on_trace = 0.5*ones(1,N_track);
        x_trace = zeros(1,N_track);
        N_off_on_trace = zeros(1,N_track);
        N_on_off_trace = zeros(1,N_track);
        t_on_trace = zeros(1,N_track);
        t_off_trace = zeros(1,N_track);
        X = 0;
        check_bad = 0;
        check_bad_prev = 0;
        check_bad_string = 'check not bad';
        lev(l).neu(n).on_off_count = 0;
        lev(l).neu(n).off_on_count = 0;
        lev(l).neu(n).input_count = 0;
        lev(l).neu(n).t_bad = 2;
        for t = 2:T
            t_mod = mod(t-1,N_track) + 1;
            t_mod_m1 = mod(t-2,N_track) + 1;
            t_mod_m10000 = mod(t-5000,N_track) + 1;
            if (l == 2) 
                % evolve HMM
                if (X == 0)
                    X = (rand(1) > level(l).neuron(n).a(1,1));
                else
                    X = (rand(1) > level(l).neuron(n).a(2,1));
                end
                level(l).neuron(n).x_k_trace(t_mod) = X;
                for s = 1:level(l).neuron(n).N_connects
                    if (X == 0)
                        syn(s).ykt = ( rand(1) > level(1).input(level(l).neuron(n).ind_connect(s)).b(1,1) );
                    else
                        syn(s).ykt = ( rand(1) > level(1).input(level(l).neuron(n).ind_connect(s)).b(2,1) );
                    end
                    level(1).input(level(l).neuron(n).ind_connect(s)).y_k_trace(t_mod) = syn(s).ykt;
                end
            elseif (l > 2)
                level(l).neuron(n).x_k_trace(t_mod) = level(l-1).neuron(n).x_k_trace(t_mod); % since state of neuron 1 and 2 are the same.
                for s = 1:level(l).neuron(n).N_connects
                    syn(s).ykt = level(l-1).neuron(level(l).neuron(n).ind_connect(s)).y_k_trace(t);
                end
            end
            % estimate parameters
            if (t_mod_m10000 < t_mod_m1)
                p_ind10000 = [ t_mod_m10000:t_mod_m1 ];
            else
                p_ind10000 = [ 1:t_mod_m1 t_mod_m10000:N_track ];
            end
            p_mx = max(p_on_trace(p_ind10000));
            p_mn = min(p_on_trace(p_ind10000));
            p_up(t_mod) = p_mn + p_up_sc*(p_mx-p_mn);
            p_down(t_mod) = p_mn + p_down_sc*(p_mx-p_mn);
            L_up(t_mod) = log((p_up(t_mod))/(1 - p_up(t_mod)));
            L_lo(t_mod) = log((p_down(t_mod))/(1 - p_down(t_mod)));
                        
            p_on = exp(L(t_mod_m1))/(1+exp(L(t_mod_m1)));
            x_pr = x_c;
            N_on_off = (1-eta)*N_on_off;
            N_off_on = (1-eta)*N_off_on;
            if (p_on > p_up(t_mod))
                x_c = 1;
                if (x_pr == 0)
                    N_off_on = eta + N_off_on;
                end
            elseif (p_on < p_down(t_mod))
                x_c = 0;
                if (x_pr == 1)
                    N_on_off = eta + N_on_off;
                end
            end
            level(l).neuron(n).x_k_est_trace(t_mod) = x_c;
            t_on = eta*(x_c + 10^(-15)) + (1-eta)*t_on;
            for s = 1:level(l).neuron(n).N_connects
                syn(s).N_on = eta*syn(s).ykt*x_c + (1-eta)*syn(s).N_on;
                syn(s).N = eta*syn(s).ykt + (1-eta)*syn(s).N;
            end
                                   
            if (t > T_stat) % keep parameters fixed for first T_stat time steps for phi to stabilise
                ronhat = max(N_off_on/(dt*(1 - t_on)),0.0001);
                roffhat = max(N_on_off/(dt*t_on),0.0001);
                for s = 1:level(l).neuron(n).N_connects
                    syn(s).qion = syn(s).N_on/(dt*t_on) + 10^(-15);
                    syn(s).qioff = (syn(s).N - syn(s).N_on)/(dt*(1 - t_on)) + 10^(-15);
                end
            end
            level(l).neuron(n).r_hat_trace(:,t_mod) = [ ronhat ; roffhat ];
            for s = 1:level(l).neuron(n).N_connects
                level(l).neuron(n).syn(s).q_hat_trace(:,t_mod) = [ syn(s).qion ; syn(s).qioff ];
            end
            p_on_trace(t_mod) = p_on;
            x_trace(t_mod) = x_c;
            N_off_on_trace(t_mod) = N_off_on;
            N_on_off_trace(t_mod) = N_on_off;
            t_on_trace(t_mod) = t_on;
            t_off_trace(t_mod) = 1 - t_on;
                       
            % display EM parameter estimates:
            if (mod(t,step_disp) == 0) | (t == T)
                    tsf = toc/3600;
                    disp([ 'EM t step ' num2str(t) ' crit_p ' num2str(crit_p) ' time so far (hrs) ' num2str(tsf) ])
            end
            if (show_figs == 1) 
                if ((mod(t,step_disp) == 0) | (t == T)) %| (check_bad == 1)
                    pause(1)
                    clf
                    subplot(4,4,1)
                    hold on
                    plot(level(l).neuron(n).r_hat_trace(1,:),'k')
                    plot(level(l).neuron(n).HMM_plot_r_on(1,:))
                    hold off
                    ylabel('r_{on}')
                    subplot(4,4,2)
                    hold on
                    plot(level(l).neuron(n).r_hat_trace(2,:),'k')
                    plot(level(l).neuron(n).HMM_plot_r_off(1,:))
                    hold off
                    ylabel('r_{off}')
                    subplot(4,4,3)
                    hold on
                    plot(L,'k')
                    plot(G,'b')
                    hold off
                    ylabel('L,G')
                    subplot(4,4,4)
                    hold on
                    plot(L,'k')
                    plot(L_true,'b')
                    plot(L_up,'r')
                    plot(L_lo,'g')
                    hold off
                    ylabel('L,L_{true},L_sm')
                    subplot(4,4,8)
                    hold on
                    plot(p_on_trace,'k')
                    plot(x_trace+1.05,'b')
                    plot(level(l).neuron(n).x_k_trace+2*1.05,'r')
                    if (l == 2)
                        plot(level(l).neuron(n).y_k_trace((t-N_track+1):t)+3*1.05,'g')
                    elseif (l > 2)
                        plot(level(l).neuron(n).y_k_trace+3*1.05,'g')
                    end
                    hold off
                    ylabel('p:k,x:b,x_{tr}:r,y:g')
                    subplot(4,4,11)
                    hold on
                    plot(N_off_on_trace,'k')
                    hold off
                    ylabel('N_off_on')
                    subplot(4,4,12)
                    hold on
                    plot(N_on_off_trace,'k')
                    hold off
                    ylabel('N_on_off')
                    subplot(4,4,15)
                    hold on
                    plot(t_on_trace,'k')
                    hold off
                    ylabel('t_on')
                    subplot(4,4,16)
                    hold on
                    plot(t_off_trace,'k')
                    hold off
                    ylabel('t_off')
                    for s = 1:min(level(l).neuron(n).N_connects,3)
                        subplot(4,4,4*s+1)
                        hold on
                        plot(level(l).neuron(n).syn(s).q_hat_trace(1,:),'k')
                        if (l == 2)
                            plot(level(l).neuron(n).syn(s).HMM_plot_q_on(1,:))
                        end
                        hold off
                        ylabel('q_{on}')
                        subplot(4,4,4*s+2)
                        hold on
                        plot(level(l).neuron(n).syn(s).q_hat_trace(2,:),'k')
                        if (l == 2)
                            plot(level(l).neuron(n).syn(s).HMM_plot_q_off(1,:))
                        end
                        hold off
                        ylabel('q_{off}')
                    end
                    pause(1)
                end
            end % show_figs
                      
            % update bayesian neuron
            % calculate neuron
            dL_spikes = 0;
            for s = 1:level(l).neuron(n).N_connects
                dL_spikes = dL_spikes + syn(s).ykt.*log(syn(s).qion/syn(s).qioff) + (1-syn(s).ykt).*log((1-syn(s).qion*dt)/(1-syn(s).qioff*dt));
            end
            L(t_mod) = L(t_mod_m1) + log(1+dt*(-roffhat+ronhat*exp(-L(t_mod_m1)))) - log(1+dt*(-ronhat+roffhat*exp(L(t_mod_m1)))) + dL_spikes;
            L_sm(t_mod) = eta_sm*L(t_mod) + (1-eta_sm)*L_sm(t_mod_m1);
            if (l == 2)
                dL_true_spikes = 0;
                for s = 1:level(l).neuron(n).N_connects            
                    dL_true_spikes = dL_true_spikes + syn(s).ykt.*log(syn(s).qion_true/syn(s).qioff_true) + (1-syn(s).ykt).*log((1-syn(s).qion_true*dt)/(1-syn(s).qioff_true*dt));
                end
                L_true(t_mod) = L_true(t_mod_m1) + log(1+dt*(-rofftrue+rontrue*exp(-L_true(t_mod_m1)))) - log(1+dt*(-rontrue+rofftrue*exp(L_true(t_mod_m1)))) + dL_true_spikes;                   
                
                dL_true_flip_spikes = 0;
                for s = 1:level(l).neuron(n).N_connects            
                    dL_true_flip_spikes = dL_true_flip_spikes + syn(s).ykt.*log(syn(s).qioff_true/syn(s).qion_true) + (1-syn(s).ykt).*log((1-syn(s).qioff_true*dt)/(1-syn(s).qion_true*dt));
                end
                L_true_flip(t_mod) = L_true_flip(t_mod_m1) + log(1+dt*(-rontrue+rofftrue*exp(-L_true_flip(t_mod_m1)))) - log(1+dt*(-rofftrue+rontrue*exp(L_true_flip(t_mod_m1)))) + dL_true_flip_spikes;                   
                                
                G(t_mod) = G(t_mod_m1) + log(1+dt*(-roffhat+ronhat*exp(-G(t_mod_m1)))) - log(1+dt*(-ronhat+roffhat*exp(G(t_mod_m1)))) + g_o*level(l).neuron(n).y_k_trace(t-1);
                if (L(t_mod) > G(t_mod) + g_o/2)
                    level(l).neuron(n).y_k_trace(t) = 1; %otherwise leaves value as zero
                %else
                %    level(l).neuron(n).y_k_trace(t) = 0;
                end
            else
                G(t_mod) = G(t_mod_m1) + log(1+dt*(-roffhat+ronhat*exp(-G(t_mod_m1)))) - log(1+dt*(-ronhat+roffhat*exp(G(t_mod_m1)))) + g_o*level(l).neuron(n).y_k_trace(t_mod_m1);
                if (L(t_mod) > G(t_mod) + g_o/2)
                    level(l).neuron(n).y_k_trace(t_mod) = 1; %otherwise leaves value as zero
                else
                    level(l).neuron(n).y_k_trace(t_mod) = 0;
                end 
            end
           
        end % end t loop
        lev(l).neu(n).check_bad = check_bad;
        lev(l).neu(n).check_bad_string = check_bad_string;
        
        ind_cp = mod([(t-step_crit+1):t]-1,N_track) + 1;
        level(l).neuron(n).ronbar = mean(level(l).neuron(n).r_hat_trace(1,ind_cp),2);
        level(l).neuron(n).roffbar = mean(level(l).neuron(n).r_hat_trace(2,ind_cp),2);
        crit_p = sum( (max(level(l).neuron(n).r_hat_trace(:,ind_cp),[],2)-min(level(l).neuron(n).r_hat_trace(:,ind_cp),[],2)) <= crit_perc*mean(level(l).neuron(n).r_hat_trace(:,ind_cp),2) );
        for s = 1:level(l).neuron(n).N_connects
            level(l).neuron(n).syn(s).qonbar = mean(level(l).neuron(n).syn(s).q_hat_trace(1,ind_cp),2);
            level(l).neuron(n).syn(s).qoffbar = mean(level(l).neuron(n).syn(s).q_hat_trace(2,ind_cp),2);
            crit_p = crit_p + sum( (max(level(l).neuron(n).syn(s).q_hat_trace(:,ind_cp),[],2)-min(level(l).neuron(n).syn(s).q_hat_trace(:,ind_cp),[],2)) <= crit_perc*mean(level(l).neuron(n).syn(s).q_hat_trace(:,ind_cp),2) );
        end
        crit_p = 2 + 2*level(l).neuron(n).N_connects - crit_p
        if (crit_p == 0) 
           level(l).neuron(n).crit_reached = 1;
        else
           level(l).neuron(n).crit_reached = 0; 
        end
        level(l).neuron(n).tot_EM_steps = t;
        
        level(l).neuron(n).crit_p = crit_p;
        t_2 = toc;
        level(l).neuron(n).tot_EM_time_mins = (t_2 - t_1)/60;
        
        % store parameters for saving
        level(l).neuron(n).r_on_hat = ronhat;
        level(l).neuron(n).r_off_hat = roffhat;
        level(l).neuron(n).th = 0;
        for s = 1:level(l).neuron(n).N_connects
            level(l).neuron(n).syn(s).q_i_on_hat = syn(s).qion;
            level(l).neuron(n).syn(s).q_i_off_hat = syn(s).qioff;
            level(l).neuron(n).syn(s).weight = log(level(l).neuron(n).syn(s).q_i_on_hat/level(l).neuron(n).syn(s).q_i_off_hat); 
            level(l).neuron(n).th = level(l).neuron(n).th + level(l).neuron(n).syn(s).q_i_on_hat - level(l).neuron(n).syn(s).q_i_off_hat;
        end     
        level(l).neuron(n).L = L;
        if (l == 2)
            level(l).neuron(n).L_true = L_true;
            level(l).neuron(n).L_true_flip = L_true_flip;
        end
        if (show_figs == 1)
            figure('position',[ 30 50 900 600 ])
            hold on
            plot(L,'k')
            plot(L_up,':r')
            plot(L_lo,':g')
            plot(G,'b')
            if (l == 2)
                tmp_trace = level(l).neuron(n).y_k_trace((t-N_track+1):t)+1.05*max([ L(:) ; G(:) ]);
            else
                tmp_trace = level(l).neuron(n).y_k_trace+1.05*max([ L(:) ; G(:) ]);                
            end
            plot(tmp_trace,'r')
            plot(level(l).neuron(n).x_k_trace+1.05*max(tmp_trace(:)),'g')
            set(gca,'fontsize',15)
            ylabel('log odds ratio')
            xlabel('time steps (dt = 0.1 ms)')
            title([ 'Output for: level ' num2str(l) ', neuron ' num2str(n) ])
            legend('L_t','L_u','L_l','G_t','O_t','x_t')
            hold off 
        end
        time_run = toc/3600
    end % for n = 1:level(l).N_neurons
end % for l = 2:N_levels   


% check HMM statistics

for l = 2:N_levels
    for n = 1:level(l).N_neurons
        disp([ 'level ' num2str(l) ', neuron ' num2str(n) ])
        ind_0 = find(level(l).neuron(n).x_k_trace(1:(end-1)) == 0);
        ind_1 = find(level(l).neuron(n).x_k_trace(1:(end-1)) == 1);
        ind_01 = find((level(l).neuron(n).x_k_trace(2:end) - level(l).neuron(n).x_k_trace(1:(end-1))) == 1);
        ind_10 = find((level(l).neuron(n).x_k_trace(2:end) - level(l).neuron(n).x_k_trace(1:(end-1))) == -1);
        ind_00 = find((level(l).neuron(n).x_k_trace(2:end) + level(l).neuron(n).x_k_trace(1:(end-1))) == 0);
        ind_11 = find((level(l).neuron(n).x_k_trace(2:end) + level(l).neuron(n).x_k_trace(1:(end-1))) == 2);
        c_a1_01 = numel(ind_01);
        c_a1_00 = numel(ind_00);
        c_a1_10 = numel(ind_10);
        c_a1_11 = numel(ind_11);
        lev(l).neu(n).r_on_sim = c_a1_01/((c_a1_00+c_a1_01)*dt);
        lev(l).neu(n).r_off_sim = c_a1_10/((c_a1_10+c_a1_11)*dt);
        a_sim = [ c_a1_00/(c_a1_00+c_a1_01) c_a1_01/(c_a1_00+c_a1_01) ; c_a1_10/(c_a1_10+c_a1_11) c_a1_11/(c_a1_10+c_a1_11) ]
        a_true = level(l).neuron(n).a
        a_hat_final = [ (1-level(l).neuron(n).r_on_hat*dt) level(l).neuron(n).r_on_hat*dt ; level(l).neuron(n).r_off_hat*dt (1-level(l).neuron(n).r_off_hat*dt) ]
        for s = 1:level(l).neuron(n).N_connects
            disp([ 'synapse ' num2str(s) ])
            if (l == 2)
                ind_01 = find(level(1).input(level(l).neuron(n).ind_connect(s)).y_k_trace(ind_0) == 1);
                ind_10 = find(level(1).input(level(l).neuron(n).ind_connect(s)).y_k_trace(ind_1) == 0);
                ind_00 = find(level(1).input(level(l).neuron(n).ind_connect(s)).y_k_trace(ind_0) == 0);
                ind_11 = find(level(1).input(level(l).neuron(n).ind_connect(s)).y_k_trace(ind_1) == 1);
                c_b1_01 = numel(ind_01);
                c_b1_00 = numel(ind_00);
                c_b1_10 = numel(ind_10);
                c_b1_11 = numel(ind_11);
                lev(l).neu(n).syn(s).q_i_on_sim = c_b1_11/((c_b1_10+c_b1_11)*dt);
                lev(l).neu(n).syn(s).q_i_off_sim = c_b1_01/((c_b1_00+c_b1_01)*dt);
                b_sim = [ c_b1_00/(c_b1_00+c_b1_01) c_b1_01/(c_b1_00+c_b1_01) ; c_b1_10/(c_b1_10+c_b1_11) c_b1_11/(c_b1_10+c_b1_11) ]
                b_true = level(1).input(level(l).neuron(n).ind_connect(s)).b
            end
            b_hat_final = [ (1-level(l).neuron(n).syn(s).q_i_off_hat*dt) level(l).neuron(n).syn(s).q_i_off_hat*dt ; (1-level(l).neuron(n).syn(s).q_i_on_hat*dt) level(l).neuron(n).syn(s).q_i_on_hat*dt ]
        end
    end
end

end_t = toc;
tot_t_hrs = end_t/3600

if ((run_type >= 2) & (run_type <= 4))
    lev(2).neu(1).r_on_true = level(2).neuron(1).r_on;
    lev(2).neu(1).r_off_true = level(2).neuron(1).r_off;
    lev(3).neu(1).r_on_true = level(3).neuron(1).r_on;
    lev(3).neu(1).r_off_true = level(3).neuron(1).r_off;
    for s = 1:level(1).N_inputs
        lev(2).neu(1).syn(s).q_i_on_true = level(1).input(s).lambda_on; 
        lev(2).neu(1).syn(s).q_i_off_true = level(1).input(s).lambda_off;
    end
end
for l = 2:N_levels
    lev(l).N_neurons = level(l).N_neurons;
    for n = 1:level(l).N_neurons
        lev(l).neu(n).r_on_init = level(l).neuron(n).r_on_init;
        lev(l).neu(n).r_off_init = level(l).neuron(n).r_off_init;
        for s = 1:level(l).neuron(n).N_connects
            lev(l).neu(n).syn(s).q_i_on_init = level(l).neuron(n).syn(s).q_i_on_init; 
            lev(l).neu(n).syn(s).q_i_off_init = level(l).neuron(n).syn(s).q_i_off_init;
        end
        level(l).neuron(n).y_k_trace = [];
        lev(l).neu(n).crit_reached = level(l).neuron(n).crit_reached; 
        lev(l).neu(n).crit_p = level(l).neuron(n).crit_p;
        lev(l).neu(n).R_on = level(l).neuron(n).r_on_hat;
        lev(l).neu(n).R_off = level(l).neuron(n).r_off_hat;
        lev(l).neu(n).R_on_bar = level(l).neuron(n).ronbar;
        lev(l).neu(n).R_off_bar = level(l).neuron(n).roffbar;
        qonh_array = zeros(1,level(l).neuron(n).N_connects);
        qoffh_array = zeros(1,level(l).neuron(n).N_connects);
        qonb_array = zeros(1,level(l).neuron(n).N_connects);
        qoffb_array = zeros(1,level(l).neuron(n).N_connects);
        for s = 1:level(l).neuron(n).N_connects
            qonh_array(s) = level(l).neuron(n).syn(s).q_i_on_hat;
            qoffh_array(s) = level(l).neuron(n).syn(s).q_i_off_hat;
            qonb_array(s) = level(l).neuron(n).syn(s).qonbar;
            qoffb_array(s) = level(l).neuron(n).syn(s).qoffbar;
        end
        lev(l).neu(n).q_i_on = qonh_array;
        lev(l).neu(n).q_i_off = qoffh_array;
        lev(l).neu(n).q_i_on_bar = qonb_array;
        lev(l).neu(n).q_i_off_bar = qoffb_array;
        lev(l).neu(n).L = level(l).neuron(n).L;
        if (l == 2)
           lev(l).neu(n).L_true = level(l).neuron(n).L_true;
           lev(l).neu(n).L_true_flip = level(l).neuron(n).L_true_flip;
        end
        lev(l).neu(n).x_k_trace = level(l).neuron(n).x_k_trace;
        lev(l).neu(n).x_k_est_trace = level(l).neuron(n).x_k_est_trace;
        lev(l).neu(n).tot_EM_time_mins = level(l).neuron(n).tot_EM_time_mins;
        lev(l).neu(n).tot_EM_steps = level(l).neuron(n).tot_EM_steps;
    end
end

eval([ 'save ' save_path 'BN_2neuron_s' num2str(seed) '_rs' num2str(rand_seed) '_p' num2str(perturb) '_s' num2str(sample_select) '_r' num2str(rate_select) '.mat lev tot_t_hrs -mat' ])

