function BN_MLEM_2Neuron(seed)

% Simulates the online maximum-likelihood-expectation-maximisation (ML-EM) algorithm in a 2 Bayesian Spiking Neuron (BSN) 
% network with N inputs going to the 1st neuron and the 1st neuron providing the only
% input to the 2nd neuron.

% This function is similar in structure to 'BN_FL_2Neuron.m' the only
% difference being the use of the MLEM algorithm instead of FL. Therefore 
% apart from the commented section on the ML-EM algorithm code segment in this file (lines 432 to 517) 
% please refer to the comments used in 'BN_FL_2Neuron.m' to understand the other the variables and code segments. 

% inputs: 
% 'seed': seeds the randomisation of the true parameter values for the
% transition and observation rates. 

% outputs: There are no outputs, however, there are options to create simulation 
% figures by setting 'show_figs' to 1, and simulated variables are automatically saved to a .mat file 

% In this function the parameter estimates are initially set to multiplicatively scaled values of the true
% parameters. This is controlled by the selection of the 'run_type' variable.

% Note rates are given in units of spikes/(0.1 ms). Therefore be aware of the following
% conversion: r = 1 spikes/s = 0.001 spikes/ms = 0.001 spikes/ 10 x 0.1 ms = 0.0001 spikes/(0.1 ms)

% This code includes additional code/checks/options that can be removed for
% efficiency purposes but have been left here. The preference has been to
% just give adequate comments. Follow the comments in a sequential order to
% make the most sense of the code. It is recommended to make your own simplified code
% but to use this as a guide.

% For help email: levink@unimelb.edu.au

% Copyright 15/11/2013 Levin Kuhlmann

tic;

show_figs = 0;
run_type = 3; %1 or 2 or 3 or 4
use_cnvg = 1;

if (run_type == 1) % single set of true parameters, run each job for 1,000,000 time steps and 10 different perturbations and 100 trials each perturbation
save_path = 'online2_data/'
rand_seed = mod(seed-1,100) + 1 % 100 random seeds
perturb = mod(ceil(seed/100)-1,10) + 1 % 10 perturbation windows
rate_select = mod(ceil(seed/1000)-1,2) + 1 % 2 sample number factors
sample_select = ceil(seed/2000) % 2 rate factors
T_steps = 1000000;
true_init = 0;
elseif (run_type == 3) % run each job for 1,000,000 time steps and 1000 different true parameter values between 0 to 0.1 and 
% using initial parameter perturbations within x to y percent
save_path = 'on_em_data_rand/'
rand_seed = seed % 1000 random seeds
perturb = 1%2%2 % 10 perturbation windows
rate_select = 1 % 2 sample number factors
sample_select = 1 % 2 rate factors
T_steps = 1000000;
true_init = 0;
end

% initialise
N_levels = 3; % number of neuron layers/levels. includes input layer
dt = .1; % ms
if (sample_select == 1)
    eta = 1/100000;
elseif (sample_select == 2)
    eta = 10^(-6);
end 

T_stat = 100;%1*10^5;
rand('state',rand_seed)
N_track = 100000;
step_disp = 100000;
use_crit_LL = 1;
if (use_crit_LL == 1)
    stop_criterion = 0.00005;
elseif (use_crit_LL == 0)
    stop_criterion = 0;
end
crit_perc = 0.01;
perc_low = [ 0 0.2 0.4 0.6 0.8 1.0 1.2 1.4 1.6 1.8 ];
perc_upp = [ 0.2 0.4 0.6 0.8 1.0 1.2 1.4 1.6 1.8 2.0 ];
perc_sc = [ 2 5 1.2 1 ];
rate_factors = [ 1 10 ];
sample_factors = [ 1 10 ];
step_crit = 100000;
perfect_int = 0; % if 0, code uses forgetting with eta. This is used in the paper. if 1, code does not use forgetting

T = T_steps*sample_factors(sample_select)%5*10^5; % number of time steps
g_o = 1.45;
level(3).N_neurons = 1;
level(2).N_neurons = 1;
level(1).N_inputs = 20;
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        if (l == 2)
            N_tmp = level(1).N_inputs;
        else
            N_tmp = level(l).N_neurons;
        end
        level(l).neuron(n).connect = ones(1,N_tmp); % determines connections to previous layer
        level(l).neuron(n).ind_connect = find(level(l).neuron(n).connect == 1);
        level(l).neuron(n).N_connects = sum(level(l).neuron(n).connect); % number of input synapses
        level(l).neuron(n).q = [ 0.5 ; 0.5 ]; % q corresponds to pi the initial conditions for the state probabilities
    end
end
% r*dt and lambda*dt need to be less than 1, make sure rates are large
% enough or T is large enough to get enough statistics
if (run_type ~= 3)
    level(3).neuron(1).r_on = 0.001*rate_factors(rate_select);%500;
    level(3).neuron(1).r_off = 0.01*rate_factors(rate_select);%1000;
    level(2).neuron(1).r_on = 0.001*rate_factors(rate_select);%500;
    level(2).neuron(1).r_off = 0.01*rate_factors(rate_select);%1000;
elseif (run_type == 3)
    level(2).neuron(1).r_off = (0.001+rand(1)*0.114)*rate_factors(rate_select);%1000;
    level(2).neuron(1).r_on = (0.001+rand(1)*0.114)*rate_factors(rate_select);
    level(3).neuron(1).r_on = level(2).neuron(1).r_on;%500;
    level(3).neuron(1).r_off = level(2).neuron(1).r_off;%1000;
    true_r_on = level(2).neuron(1).r_on
    true_r_off = level(2).neuron(1).r_off
end
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        level(l).neuron(n).a = [ (1-level(l).neuron(n).r_on*dt) level(l).neuron(n).r_on*dt ; level(l).neuron(n).r_off*dt (1-level(l).neuron(n).r_off*dt) ];
    end
end
for s = 1:level(1).N_inputs
    if (run_type ~= 3) 
        level(1).input(s).lambda_on = 0.05*rate_factors(rate_select);%.08*1; 
        level(1).input(s).lambda_off = 0.01*rate_factors(rate_select);
    elseif (run_type == 3)
        level(1).input(s).lambda_on = (0.001+rand(1)*0.999)*rate_factors(rate_select);%.08*1;
        level(1).input(s).lambda_off = (0.001+rand(1)*0.999)*rate_factors(rate_select);
        true_q_on = level(1).input(s).lambda_on
        true_q_off = level(1).input(s).lambda_off
    end
    level(1).input(s).b = [ (1-level(1).input(s).lambda_off*dt) level(1).input(s).lambda_off*dt ; (1-level(1).input(s).lambda_on*dt) level(1).input(s).lambda_on*dt ];
end

if (use_cnvg == 1)
    cnvg.r_on_follow = zeros(1,T_steps);
    cnvg.r_off_follow = zeros(1,T_steps);
    cnvg.q_on_follow = zeros(level(1).N_inputs,T_steps);
    cnvg.q_off_follow = zeros(level(1).N_inputs,T_steps);
    cnvg.r_on = level(2).neuron(1).r_on;
    cnvg.r_off = level(2).neuron(1).r_off;
    for s = 1:level(1).N_inputs
        cnvg.q_on(s) = level(1).input(s).lambda_on;
        cnvg.q_off(s) = level(1).input(s).lambda_off;
    end
end    

% create EM variables for each neuron
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        level(l).neuron(n).a_hat_trace = NaN(2,2,N_track);
        for s = 1:level(l).neuron(n).N_connects
            level(l).neuron(n).syn(s).b_hat_trace = NaN(2,2,N_track);
        end
        level(l).neuron(n).r_on_hat = [];
        level(l).neuron(n).r_off_hat = [];
    end
end

% initialise the first values of q, a_hat and
% syn(i).lambda_on_hat, syn(i).lambda_off_hat, gamma, and phi
if (true_init == 0) % initialise with random values
    for l = 2:N_levels
        for n = 1:level(l).N_neurons
            rate_min = 1*10^(-6);
            ron_percl = perc_low(perturb)*level(l).neuron(n).r_on;
            ron_percu = perc_upp(perturb)*level(l).neuron(n).r_on;
            ron_t1 = level(l).neuron(n).r_on + ron_percl;
            ron_t2 = level(l).neuron(n).r_on + ron_percu;
            ron_b2 = max(level(l).neuron(n).r_on - ron_percl,rate_min);
            ron_b1 = max(level(l).neuron(n).r_on - ron_percu,rate_min);
            ratio_tb = (ron_t2-ron_t1)/(ron_t2-ron_t1+ron_b2-ron_b1);
            rndc = rand(1,1);
            if (rndc <= ratio_tb)
                rontr = ron_t1 + (ron_t2 - ron_t1)*rndc/ratio_tb;
            else
                rontr = ron_b1 + (ron_b2 - ron_b1)*(1-rndc)/(1-ratio_tb);                
            end
            roff_percl = perc_low(perturb)*level(l).neuron(n).r_off;
            roff_percu = perc_upp(perturb)*level(l).neuron(n).r_off;
            roff_t1 = level(l).neuron(n).r_off + roff_percl;
            roff_t2 = level(l).neuron(n).r_off + roff_percu;
            roff_b2 = max(level(l).neuron(n).r_off - roff_percl,rate_min);
            roff_b1 = max(level(l).neuron(n).r_off - roff_percu,rate_min);
            ratio_tb = (roff_t2-roff_t1)/(roff_t2-roff_t1+roff_b2-roff_b1);
            rndc = rand(1,1);
            if (rndc <= ratio_tb)
                rofftr = roff_t1 + (roff_t2 - roff_t1)*rndc/ratio_tb;
            else
                rofftr = roff_b1 + (roff_b2 - roff_b1)*(1-rndc)/(1-ratio_tb);                
            end
            
            if (run_type == 3)
                rontr = perc_sc(perturb)*level(l).neuron(n).r_on;
                rofftr = perc_sc(perturb)*level(l).neuron(n).r_off;
            end

            level(l).neuron(n).a_hat_trace(:,:,1) = [ (1-rontr*dt) rontr*dt ; rofftr*dt (1-rofftr*dt) ];
            level(l).neuron(n).r_on_init = rontr;
            level(l).neuron(n).r_off_init = rofftr;
            
            if (l == 2)
                qon_rate_min = 1*10^(-6);
                qoff_rate_min = 1*10^(-6);
            elseif (l == 3)
                qon_rate_min = 0.03;
                qoff_rate_min = 1*10^(-6);
            end
            
            for s = 1:level(l).neuron(n).N_connects
                if (l == 2)
                    lon_percl = perc_low(perturb)*level(1).input(s).lambda_on;
                    lon_percu = perc_upp(perturb)*level(1).input(s).lambda_on;
                    lon_t1 = level(1).input(s).lambda_on + lon_percl;
                    lon_t2 = level(1).input(s).lambda_on + lon_percu;
                    lon_b2 = max(level(1).input(s).lambda_on - lon_percl,qon_rate_min);
                    lon_b1 = max(level(1).input(s).lambda_on - lon_percu,qon_rate_min);
                else
                    lon_percl = perc_low(perturb)*0.3879;
                    lon_percu = perc_upp(perturb)*0.3879;
                    lon_t1 = 0.3879 + lon_percl;
                    lon_t2 = 0.3879 + lon_percu;
                    lon_b2 = max(0.3879 - lon_percl,qon_rate_min);
                    lon_b1 = max(0.3879 - lon_percu,qon_rate_min);
                end
                lon_b1 = max(level(1).input(s).lambda_on - lon_percu,qon_rate_min);
                ratio_tb = (lon_t2-lon_t1)/(lon_t2-lon_t1+lon_b2-lon_b1);
                rndc = rand(1,1);
                if (rndc <= ratio_tb)
                    lon_rnd = lon_t1 + (lon_t2 - lon_t1)*rndc/ratio_tb;
                else
                    lon_rnd = lon_b1 + (lon_b2 - lon_b1)*(1-rndc)/(1-ratio_tb);                
                end
                if (l == 2)
                    loff_percl = perc_low(perturb)*level(1).input(s).lambda_off;
                    loff_percu = perc_upp(perturb)*level(1).input(s).lambda_off;
                    loff_t1 = level(1).input(s).lambda_off + loff_percl;
                    loff_t2 = level(1).input(s).lambda_off + loff_percu;
                    loff_b2 = max(level(1).input(s).lambda_off - loff_percl,qoff_rate_min);
                    loff_b1 = max(level(1).input(s).lambda_off - loff_percu,qoff_rate_min);
                else
                    loff_percl = perc_low(perturb)*7.5285*10^-3;
                    loff_percu = perc_upp(perturb)*7.5285*10^-3;
                    loff_t1 = 7.5285*10^-3 + loff_percl;
                    loff_t2 = 7.5285*10^-3 + loff_percu;
                    loff_b2 = max(7.5285*10^-3 - loff_percl,qoff_rate_min);
                    loff_b1 = max(7.5285*10^-3 - loff_percu,qoff_rate_min);
                end                    
                loff_b1 = max(level(1).input(s).lambda_off - loff_percu,qoff_rate_min);
                ratio_tb = (loff_t2-loff_t1)/(loff_t2-loff_t1+loff_b2-loff_b1);
                rndc = rand(1,1);
                if (rndc <= ratio_tb)
                    loff_rnd = loff_t1 + (loff_t2 - loff_t1)*rndc/ratio_tb;
                else
                    loff_rnd = loff_b1 + (loff_b2 - loff_b1)*(1-rndc)/(1-ratio_tb);                
                end
                if (run_type == 3)
                    if (l == 2)
                        lon_rnd = perc_sc(perturb)*level(1).input(s).lambda_on;
                        loff_rnd = perc_sc(perturb)*level(1).input(s).lambda_off;
                    else
                        lon_rnd = perc_sc(perturb)*level(1).input(s).lambda_on;
                        loff_rnd = perc_sc(perturb)*level(1).input(s).lambda_off;
                    end
                end

                level(l).neuron(n).syn(s).b_hat_trace(:,:,1) = [ (1-loff_rnd*dt) loff_rnd*dt ; (1-lon_rnd*dt) lon_rnd*dt ];
                level(l).neuron(n).syn(s).q_i_on_init = lon_rnd;
                level(l).neuron(n).syn(s).q_i_off_init = loff_rnd;
            end
        end
    end
elseif (true_init == 1) % initialise with true parameter values
    for l = 2:N_levels
        for n = 1:level(l).N_neurons
            level(l).neuron(n).a_hat_trace(:,:,1) = level(l).neuron(n).a;
            level(l).neuron(n).r_on_init = level(l).neuron(n).a(1,2)/dt;
            level(l).neuron(n).r_off_init = level(l).neuron(n).a(2,1)/dt;
            if (l == 2)
                for s = 1:level(l).neuron(n).N_connects
                    level(l).neuron(n).syn(s).b_hat_trace(:,:,1) = level(1).input(level(l).neuron(n).ind_connect(s)).b;
                    level(l).neuron(n).syn(s).q_i_on_init = level(1).input(level(l).neuron(n).ind_connect(s)).b(2,2)/dt;
                    level(l).neuron(n).syn(s).q_i_off_init = level(1).input(level(l).neuron(n).ind_connect(s)).b(1,2)/dt;
                end
            else
                for s = 1:level(l).neuron(n).N_connects
                    level(l).neuron(n).syn(s).b_hat_trace(:,:,1) = level(1).input(level(l).neuron(n).ind_connect(s)).b; % these are undefined for the generative model so choose random values
                    level(l).neuron(n).syn(s).q_i_on_init = level(1).input(level(l).neuron(n).ind_connect(s)).b(2,2)/dt;
                    level(l).neuron(n).syn(s).q_i_off_init = level(1).input(level(l).neuron(n).ind_connect(s)).b(1,2)/dt;
                end
            end
        end
    end
end

% create plot data of HMM values
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        level(l).neuron(n).HMM_plot_a = repmat(level(l).neuron(n).a(:),[ 1 N_track ]);
        if (l == 2)
            for s = 1:level(l).neuron(n).N_connects
                level(l).neuron(n).syn(s).HMM_plot_b = repmat(level(1).input(level(l).neuron(n).ind_connect(s)).b(:),[ 1 N_track ]);
            end
        end
    end
end

% create arrays to store HMM simulation results and bayesian neuron outputs
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        level(l).neuron(n).x_k_trace = zeros(1,N_track); % neuron hidden state
        level(l).neuron(n).x_k_est_trace = zeros(1,N_track); % neuron hidden state
        if (l == 2)
            level(l).neuron(n).y_k_trace = zeros(1,T); % neuron output spikes
        else
            level(l).neuron(n).y_k_trace = zeros(1,N_track);
        end
    end        
end
for s = 1:level(1).N_inputs
    level(1).input(s).y_k_trace = zeros(1,N_track); % network input spikes/HMM observations
end

% EM
% run the forward backward algorithm for each level > 1 and each neuron in
% the corresponding level, run lower levels first
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        t_1 = toc;
        % use parameter estimates and numerically integrate L and G to determine output
        % spikes and run online EM.
        % initialise L and G and define relevant parameters
        level(l).neuron(n).r_on_hat = level(l).neuron(n).a_hat_trace(1,2,1)/dt;
        level(l).neuron(n).r_off_hat = level(l).neuron(n).a_hat_trace(2,1,1)/dt;
        level(l).neuron(n).th = 0;
        for s = 1:level(l).neuron(n).N_connects
            level(l).neuron(n).syn(s).q_i_on_hat = level(l).neuron(n).syn(s).b_hat_trace(2,2,1)/dt;
            level(l).neuron(n).syn(s).q_i_off_hat = level(l).neuron(n).syn(s).b_hat_trace(1,2,1)/dt;
            level(l).neuron(n).syn(s).weight = log(level(l).neuron(n).syn(s).q_i_on_hat/level(l).neuron(n).syn(s).q_i_off_hat); 
            level(l).neuron(n).th = level(l).neuron(n).th + level(l).neuron(n).syn(s).q_i_on_hat - level(l).neuron(n).syn(s).q_i_off_hat;
        end
        ronhat = level(l).neuron(n).r_on_hat;
        roffhat = level(l).neuron(n).r_off_hat;
        rontrue = level(l).neuron(n).r_on;
        rofftrue = level(l).neuron(n).r_off;
        th = level(l).neuron(n).th;
        disp('initial values for:')
        disp([ 'level ' num2str(l) ', neuron ' num2str(n) ', ronhat ' num2str(ronhat) ', roffhat ' num2str(roffhat) ', th ' num2str(th) ])
        for s = 1:level(l).neuron(n).N_connects
            wght = level(l).neuron(n).syn(s).weight;
            disp([ 'synapse ' num2str(s) ', wght ' num2str(wght) ])
            syn(s).qion = level(l).neuron(n).syn(s).q_i_on_hat;
            syn(s).qioff = level(l).neuron(n).syn(s).q_i_off_hat;
        end
        
        L = zeros(1,N_track);
        L(1) = log(level(l).neuron(n).r_on_hat/level(l).neuron(n).r_off_hat);
        G = L;
        % since all input spikes and parameters are known for neuron 2 calculate input time series for L_true.
        if (l == 2)
            L_true = zeros(1,N_track);
            L_true(1) = L(1);
            L_true_flip = zeros(1,N_track);
            L_true_flip(1) = L(1);
        end
        % initialise additional EM terms (variable names correspond closely
        % to variables in MLEM algorithm equations in the paper, namely phi, gamma and q (note 'q' here is 'Q' in the paper)
        phi = zeros(2,2,2,2,level(l).neuron(n).N_connects);
        phi_sum = phi;
        phi_end = phi;
        phi_prev = zeros(2,2,2,2,level(l).neuron(n).N_connects);
        for s = 1:level(l).neuron(n).N_connects
            syn(s).gamma = zeros(2,2);
            syn(s).q = zeros(2,1);
            temp_q = 0.5;% rand(1,1);
            syn(s).q_prev = [ temp_q ; (1 - temp_q) ];
        end
        safety_0 = 10^(-120);
        a_hat = level(l).neuron(n).a_hat_trace(:,:,1);
        for s = 1:level(l).neuron(n).N_connects
            syn(s).b_hat = level(l).neuron(n).syn(s).b_hat_trace(:,:,1);
        end
        if (show_figs == 1)
            figure('position',[ 30 50 900 600 ])
        end
        % run bayesian neuron and EM online
        level(l).neuron(n).crit_reached = 0;
        level(l).neuron(n).tot_EM_steps = NaN;
        first = 1;
        crit_p = 4 + 4*level(l).neuron(n).N_connects;
        rontrue = level(l).neuron(n).r_on;
        rofftrue = level(l).neuron(n).r_off; 
        if (l == 2)
            for s = 1:level(l).neuron(n).N_connects
                syn(s).qion_true = level(1).input(s).lambda_on;
                syn(s).qioff_true = level(1).input(s).lambda_off;
            end
        end
        X = 0;
        level(l).neuron(n).total_input_spikes = 0;
        level(l).neuron(n).total_output_spikes = 0;
        for t = 2:T
            t_mod = mod(t-1,N_track) + 1;
            t_mod_m1 = mod(t-2,N_track) + 1;
            if (l == 2) 
                % evolve HMM
                if (X == 0)
                    X = (rand(1) > level(l).neuron(n).a(1,1));
                else
                    X = (rand(1) > level(l).neuron(n).a(2,1));
                end
                level(l).neuron(n).x_k_trace(t_mod) = X;
                for s = 1:level(l).neuron(n).N_connects
                    if (X == 0)
                        syn(s).ykt = ( rand(1) > level(1).input(level(l).neuron(n).ind_connect(s)).b(1,1) );
                    else
                        syn(s).ykt = ( rand(1) > level(1).input(level(l).neuron(n).ind_connect(s)).b(2,1) );
                    end
                    level(1).input(level(l).neuron(n).ind_connect(s)).y_k_trace(t_mod) = syn(s).ykt;
                end
            elseif (l > 2) % set input spikes
                level(l).neuron(n).x_k_trace(t_mod) = level(l-1).neuron(n).x_k_trace(t_mod); % since state of neuron 1 and 2 are the same.
                for s = 1:level(l).neuron(n).N_connects
                    syn(s).ykt = level(l-1).neuron(level(l).neuron(n).ind_connect(s)).y_k_trace(t);
                end
            end
            % estimate parameters with EM
            % update gamma (using exponents of log sums to give better
            % numerical stability for small probabilities and also to
            % probably increase computational efficiency by avoiding multiplications). 
            sum_mn = zeros(1,level(l).neuron(n).N_connects);
            for s = 1:level(l).neuron(n).N_connects
                for m = 1:2
                    for nn = 1:2
                        sum_mn(s) = sum_mn(s) + exp(log(a_hat(m,nn)) + log(syn(s).b_hat(nn,syn(s).ykt+1)) + log(syn(s).q_prev(m)));
                    end
                end
            end
            for ll = 1:2
                for h = 1:2
                    for s = 1:level(l).neuron(n).N_connects
                        syn(s).gamma(ll,h) = exp(log(exp(log(a_hat(ll,h)) + log(syn(s).b_hat(h,syn(s).ykt+1)))) - log(sum_mn(s)));
                    end
                end
            end

            % determine the state probability vector for each synapse
            q1_avg = 0; % used to store the probability that the state is on averaged across the synapses. 
            % This average probability can be thresholded to generate a
            % rough state estimate.
            for s = 1:level(l).neuron(n).N_connects
                 syn(s).q = exp(log(syn(s).gamma(1,:)') + log(syn(s).q_prev(1))) + exp(log(syn(s).gamma(2,:)') + log(syn(s).q_prev(2))); % state probability vector for the current synapse s
                 q1_avg = q1_avg + syn(s).q(1);
            end
            q1_avg = q1_avg./level(l).neuron(n).N_connects; % average probability that the state is on.
            
            level(l).neuron(n).x_k_est_trace(t_mod) = (q1_avg > 0.5);
            
            if isnan(a_hat(1,1))
                dfadfasd
            end
            
            % update phi
            for s = 1:level(l).neuron(n).N_connects
                for k = 0:1
                    for i = 0:1
                        for j = 0:1
                            for h = 0:1  
                                if (perfect_int == 0) % uses forgetting with eta. This is used in the paper.
                                    phi(i+1,j+1,k+1,h+1,s) = exp(log(syn(s).gamma(1,h+1)) + log( phi_prev(i+1,j+1,k+1,1,s)+eta*( (syn(s).ykt == k)*(i == 0)*(j == h)*syn(s).q_prev(1) - phi_prev(i+1,j+1,k+1,1,s) + safety_0 ) )) + exp(log(syn(s).gamma(2,h+1)) + log( phi_prev(i+1,j+1,k+1,2,s)+eta*( (syn(s).ykt == k)*(i == 1)*(j == h)*syn(s).q_prev(2) - phi_prev(i+1,j+1,k+1,2,s) + safety_0 ) ));
                                elseif (perfect_int == 1) % does not use forgetting. This is not used in the paper.
                                    phi(i+1,j+1,k+1,h+1,s) = exp(log(syn(s).gamma(1,h+1)) + log( (syn(s).ykt == k)*(i == 0)*(j == h)*syn(s).q_prev(1) + safety_0 )) + exp(log(syn(s).gamma(2,h+1)) + log( (syn(s).ykt == k)*(i == 1)*(j == h)*syn(s).q_prev(2) + safety_0 ));                                    
                                end
                            end
                        end
                    end
                end
            end
            
            if (perfect_int == 1)
                phi_sum = phi_sum + phi;
            end
                      
            for s = 1:level(l).neuron(n).N_connects
                level(l).neuron(n).total_input_spikes = level(l).neuron(n).total_input_spikes + syn(s).ykt;
            end
            
            if (t > T_stat) % keep parameters fixed for first T_stat time steps for phi to stabilise
                % calculate most recent parameter estimates
                if (perfect_int == 0)
                    a_hat = sum(sum(sum(phi,5),4),3)./repmat(sum(sum(sum(sum(phi,5),4),3),2),[ 1 2 ]);
                else
                    a_hat = sum(sum(sum(phi_sum,5),4),3)./repmat(sum(sum(sum(sum(phi_sum,5),4),3),2),[ 1 2 ]);                    
                end
                for s = 1:level(l).neuron(n).N_connects
                    if (perfect_int == 0)
                        syn(s).b_hat = exp(log(sum(sum(permute(phi(:,:,:,:,s),[ 2 3 4 1 ]),4),3)) - log(repmat(sum(sum(sum(permute(phi(:,:,:,:,s),[ 2 3 4 1 ]),4),3),2),[ 1 2 ])));                       
                    else
                        syn(s).b_hat = exp(log(sum(sum(permute(phi_sum(:,:,:,:,s),[ 2 3 4 1 ]),4),3)) - log(repmat(sum(sum(sum(permute(phi_sum(:,:,:,:,s),[ 2 3 4 1 ]),4),3),2),[ 1 2 ])));                                               
                    end                        
                end
            end
            % buffer a_hat and b_hat
            level(l).neuron(n).a_hat_trace(:,:,t_mod) = a_hat;
            for s = 1:level(l).neuron(n).N_connects
                level(l).neuron(n).syn(s).b_hat_trace(:,:,t_mod) = syn(s).b_hat;
            end
            % shift EM variables into previous variables
            for s = 1:level(l).neuron(n).N_connects
                syn(s).q_prev = syn(s).q;
            end
            phi_prev = phi;
            
            % display EM parameter estimates:
            if (mod(t,step_disp) == 0) | (t == T)
                    tsf = toc/3600;
                    disp([ 'EM t step ' num2str(t) ' crit_p ' num2str(crit_p) ' time so far (hrs) ' num2str(tsf) ])
            end
            if (show_figs == 1)
                if (mod(t,step_disp) == 0) | (t == T)
                    pause(0.002)
                    clf
                    subplot(4,4,1)
                    hold on
                    plot(permute(level(l).neuron(n).a_hat_trace(1,1,:),[ 2 3 1]),'k')
                    plot(level(l).neuron(n).HMM_plot_a(1,:))
                    hold off
                    ylabel('a_{00}')
                    subplot(4,4,2)
                    hold on
                    plot(permute(level(l).neuron(n).a_hat_trace(1,2,:),[ 2 3 1]),'k')
                    plot(level(l).neuron(n).HMM_plot_a(3,:))
                    hold off
                    ylabel('a_{01}')
                    subplot(4,4,3)
                    hold on
                    plot(permute(level(l).neuron(n).a_hat_trace(2,1,:),[ 2 3 1]),'k')
                    plot(level(l).neuron(n).HMM_plot_a(2,:))
                    hold off
                    ylabel('a_{10}')
                    subplot(4,4,4)
                    hold on
                    plot(permute(level(l).neuron(n).a_hat_trace(2,2,:),[ 2 3 1]),'k')
                    plot(level(l).neuron(n).HMM_plot_a(4,:))
                    hold off
                    ylabel('a_{11}')
                    for s = 1:min(level(l).neuron(n).N_connects,3)
                        subplot(4,4,4*s+1)
                        hold on
                        plot(permute(level(l).neuron(n).syn(s).b_hat_trace(1,1,:),[ 2 3 1]),'k')
                        if (l == 2)
                            plot(level(l).neuron(n).syn(s).HMM_plot_b(1,:))
                        end
                        hold off
                        ylabel('b_{00}')
                        subplot(4,4,4*s+2)
                        hold on
                        plot(permute(level(l).neuron(n).syn(s).b_hat_trace(1,2,:),[ 2 3 1]),'k')
                        if (l == 2)
                            plot(level(l).neuron(n).syn(s).HMM_plot_b(3,:))
                        end
                        hold off
                        ylabel('b_{01}')
                        subplot(4,4,4*s+3)
                        hold on
                        plot(permute(level(l).neuron(n).syn(s).b_hat_trace(2,1,:),[ 2 3 1]),'k')
                        if (l == 2)
                            plot(level(l).neuron(n).syn(s).HMM_plot_b(2,:))
                        end
                        hold off
                        ylabel('b_{10}')
                        subplot(4,4,4*s+4)
                        hold on
                        plot(permute(level(l).neuron(n).syn(s).b_hat_trace(2,2,:),[ 2 3 1]),'k')
                        if (l == 2)
                            plot(level(l).neuron(n).syn(s).HMM_plot_b(4,:))
                        end
                        hold off
                        ylabel('b_{11}')
                    end
                    pause(0.002)
                end
            end % show_figs

            % run for loop over time for bayesian neuron
            % convert EM parameters into bayesian neuron parameters
                       
            % calculate neuron
            ronhat = a_hat(1,2)/dt;
            roffhat = a_hat(2,1)/dt;
            dL_spikes = 0;
            for s = 1:level(l).neuron(n).N_connects
                syn(s).qion = syn(s).b_hat(2,2)/dt;
                syn(s).qioff = syn(s).b_hat(1,2)/dt;
                dL_spikes = dL_spikes + syn(s).ykt.*log(syn(s).qion/syn(s).qioff) + (1-syn(s).ykt).*log((1-syn(s).qion*dt)/(1-syn(s).qioff*dt));
            end
            L(t_mod) = L(t_mod_m1) + log(1+dt*(-roffhat+ronhat*exp(-L(t_mod_m1)))) - log(1+dt*(-ronhat+roffhat*exp(L(t_mod_m1)))) + dL_spikes;
            if (l == 2)
                dL_true_spikes = 0;
                for s = 1:level(l).neuron(n).N_connects            
                    dL_true_spikes = dL_true_spikes + syn(s).ykt.*log(syn(s).qion_true/syn(s).qioff_true) + (1-syn(s).ykt).*log((1-syn(s).qion_true*dt)/(1-syn(s).qioff_true*dt));
                end
                L_true(t_mod) = L_true(t_mod_m1) + log(1+dt*(-rofftrue+rontrue*exp(-L_true(t_mod_m1)))) - log(1+dt*(-rontrue+rofftrue*exp(L_true(t_mod_m1)))) + dL_true_spikes;                   
                
                dL_true_flip_spikes = 0;
                for s = 1:level(l).neuron(n).N_connects            
                    dL_true_flip_spikes = dL_true_flip_spikes + syn(s).ykt.*log(syn(s).qioff_true/syn(s).qion_true) + (1-syn(s).ykt).*log((1-syn(s).qioff_true*dt)/(1-syn(s).qion_true*dt));
                end
                L_true_flip(t_mod) = L_true_flip(t_mod_m1) + log(1+dt*(-rontrue+rofftrue*exp(-L_true_flip(t_mod_m1)))) - log(1+dt*(-rofftrue+rontrue*exp(L_true_flip(t_mod_m1)))) + dL_true_flip_spikes;                   
                                
                G(t_mod) = G(t_mod_m1) + log(1+dt*(-roffhat+ronhat*exp(-G(t_mod_m1)))) - log(1+dt*(-ronhat+roffhat*exp(G(t_mod_m1)))) + g_o*level(l).neuron(n).y_k_trace(t-1);
                if (L(t_mod) > G(t_mod) + g_o/2)
                    level(l).neuron(n).y_k_trace(t) = 1; %otherwise leaves value as zero
                %else
                %    level(l).neuron(n).y_k_trace(t) = 0;
                end
            else
                G(t_mod) = G(t_mod_m1) + log(1+dt*(-roffhat+ronhat*exp(-G(t_mod_m1)))) - log(1+dt*(-ronhat+roffhat*exp(G(t_mod_m1)))) + g_o*level(l).neuron(n).y_k_trace(t_mod_m1);
                if (L(t_mod) > G(t_mod) + g_o/2)
                    level(l).neuron(n).y_k_trace(t_mod) = 1; %otherwise leaves value as zero
                else
                    level(l).neuron(n).y_k_trace(t_mod) = 0;
                end 
            end
            
            level(l).neuron(n).total_output_spikes = level(l).neuron(n).total_output_spikes + level(l).neuron(n).y_k_trace(t_mod);
            
            if (use_cnvg == 1)
                if (t_mod == N_track) & (l == 2) & (n == 1)
                    cnvg.r_on_follow((t-N_track+1):t) = permute(level(l).neuron(n).a_hat_trace(1,2,:)/dt,[ 1 3 2 ]);%level(l).neuron(n).r_hat_trace(1,:);
                    cnvg.r_off_follow((t-N_track+1):t) = permute(level(l).neuron(n).a_hat_trace(2,1,:)/dt,[ 1 3 2 ]);%level(l).neuron(n).r_hat_trace(2,:);
                    for s = 1:level(1).N_inputs
                        cnvg.q_on_follow(s,(t-N_track+1):t) = permute(level(l).neuron(n).syn(s).b_hat_trace(2,2,:)/dt,[ 1 3 2 ]);%level(l).neuron(n).syn(s).q_hat_trace(1,:);
                        cnvg.q_off_follow(s,(t-N_track+1):t) = permute(level(l).neuron(n).syn(s).b_hat_trace(1,2,:)/dt,[ 1 3 2 ]);%level(l).neuron(n).syn(s).q_hat_trace(2,:);
                    end
                end
            end
        end % end t loop
        
        ind_cp = mod([(t-step_crit+1):t]-1,N_track) + 1;
        level(l).neuron(n).ronbar = mean(permute(level(l).neuron(n).a_hat_trace(1,2,ind_cp)/dt,[ 3 1 2 ]));
        level(l).neuron(n).roffbar = mean(permute(level(l).neuron(n).a_hat_trace(2,1,ind_cp)/dt,[ 3 1 2 ]));
        crit_p = sum(sum( (max(level(l).neuron(n).a_hat_trace(:,:,ind_cp),[],3)-min(level(l).neuron(n).a_hat_trace(:,:,ind_cp),[],3)) <= crit_perc*mean(level(l).neuron(n).a_hat_trace(:,:,ind_cp),3) ));
        for s = 1:level(l).neuron(n).N_connects
            level(l).neuron(n).syn(s).qonbar = mean(level(l).neuron(n).syn(s).b_hat_trace(2,2,ind_cp)/dt,3);
            level(l).neuron(n).syn(s).qoffbar = mean(level(l).neuron(n).syn(s).b_hat_trace(1,2,ind_cp)/dt,3);
            crit_p = crit_p + sum(sum( (max(level(l).neuron(n).syn(s).b_hat_trace(:,:,ind_cp),[],3)-min(level(l).neuron(n).syn(s).b_hat_trace(:,:,ind_cp),[],3)) <= crit_perc*mean(level(l).neuron(n).syn(s).b_hat_trace(:,:,ind_cp),3) ));
        end
        crit_p = 4 + 4*level(l).neuron(n).N_connects - crit_p
        if (crit_p == 0) 
           level(l).neuron(n).crit_reached = 1;
        else
           level(l).neuron(n).crit_reached = 0; 
        end
        level(l).neuron(n).tot_EM_steps = t;
        
        level(l).neuron(n).crit_p = crit_p;
        t_2 = toc;
        level(l).neuron(n).tot_EM_time_mins = (t_2 - t_1)/60;

        level(l).neuron(n).r_on_hat = a_hat(1,2)/dt;
        level(l).neuron(n).r_off_hat = a_hat(2,1)/dt;
        level(l).neuron(n).th = 0;
        for s = 1:level(l).neuron(n).N_connects
            level(l).neuron(n).syn(s).q_i_on_hat = syn(s).b_hat(2,2)/dt;
            level(l).neuron(n).syn(s).q_i_off_hat = syn(s).b_hat(1,2)/dt;
            level(l).neuron(n).syn(s).weight = log(level(l).neuron(n).syn(s).q_i_on_hat/level(l).neuron(n).syn(s).q_i_off_hat); 
            level(l).neuron(n).th = level(l).neuron(n).th + level(l).neuron(n).syn(s).q_i_on_hat - level(l).neuron(n).syn(s).q_i_off_hat;
        end    
            
        level(l).neuron(n).L = L;
        if (l == 2)
            level(l).neuron(n).L_true = L_true;
            level(l).neuron(n).L_true_flip = L_true_flip;
        end
        if (show_figs == 1)
            figure('position',[ 30 50 900 600 ])
            hold on
            plot(L,'k')
            plot(G,'b')
            if (l == 2)
                tmp_trace = level(l).neuron(n).y_k_trace((t-N_track+1):t)+1.05*max([ L(:) ; G(:) ]);
            else
                tmp_trace = level(l).neuron(n).y_k_trace+1.05*max([ L(:) ; G(:) ]);                
            end
            plot(tmp_trace,'r')
            plot(level(l).neuron(n).x_k_trace+1.05*max(tmp_trace(:)),'g')
            set(gca,'fontsize',15)
            ylabel('log odds ratio')
            xlabel('time steps (dt = 0.1 ms)')
            title([ 'Output for: level ' num2str(l) ', neuron ' num2str(n) ])
            legend('L_t','G_t','O_t','x_t')
            hold off 
        end
        time_run = toc/3600
        
    end % for n = 1:level(l).N_neurons
end % for l = 2:N_levels   


% check HMM statistics
for l = 2:N_levels
    for n = 1:level(l).N_neurons
        disp([ 'level ' num2str(l) ', neuron ' num2str(n) ])
        ind_0 = find(level(l).neuron(n).x_k_trace(1:(end-1)) == 0);
        ind_1 = find(level(l).neuron(n).x_k_trace(1:(end-1)) == 1);
        ind_01 = find((level(l).neuron(n).x_k_trace(2:end) - level(l).neuron(n).x_k_trace(1:(end-1))) == 1);
        ind_10 = find((level(l).neuron(n).x_k_trace(2:end) - level(l).neuron(n).x_k_trace(1:(end-1))) == -1);
        ind_00 = find((level(l).neuron(n).x_k_trace(2:end) + level(l).neuron(n).x_k_trace(1:(end-1))) == 0);
        ind_11 = find((level(l).neuron(n).x_k_trace(2:end) + level(l).neuron(n).x_k_trace(1:(end-1))) == 2);
        c_a1_01 = numel(ind_01);
        c_a1_00 = numel(ind_00);
        c_a1_10 = numel(ind_10);
        c_a1_11 = numel(ind_11);
        lev(l).neu(n).r_on_sim = c_a1_01/((c_a1_00+c_a1_01)*dt);
        lev(l).neu(n).r_off_sim = c_a1_10/((c_a1_10+c_a1_11)*dt);
        a_sim = [ c_a1_00/(c_a1_00+c_a1_01) c_a1_01/(c_a1_00+c_a1_01) ; c_a1_10/(c_a1_10+c_a1_11) c_a1_11/(c_a1_10+c_a1_11) ]
        a_true = level(l).neuron(n).a
        a_hat_final = level(l).neuron(n).a_hat_trace(:,:,end)
        for s = 1:level(l).neuron(n).N_connects
            disp([ 'synapse ' num2str(s) ])
            if (l == 2)
                ind_01 = find(level(1).input(level(l).neuron(n).ind_connect(s)).y_k_trace(ind_0) == 1);
                ind_10 = find(level(1).input(level(l).neuron(n).ind_connect(s)).y_k_trace(ind_1) == 0);
                ind_00 = find(level(1).input(level(l).neuron(n).ind_connect(s)).y_k_trace(ind_0) == 0);
                ind_11 = find(level(1).input(level(l).neuron(n).ind_connect(s)).y_k_trace(ind_1) == 1);
                c_b1_01 = numel(ind_01);
                c_b1_00 = numel(ind_00);
                c_b1_10 = numel(ind_10);
                c_b1_11 = numel(ind_11);
                lev(l).neu(n).syn(s).q_i_on_sim = c_b1_11/((c_b1_10+c_b1_11)*dt);
                lev(l).neu(n).syn(s).q_i_off_sim = c_b1_01/((c_b1_00+c_b1_01)*dt);
                b_sim = [ c_b1_00/(c_b1_00+c_b1_01) c_b1_01/(c_b1_00+c_b1_01) ; c_b1_10/(c_b1_10+c_b1_11) c_b1_11/(c_b1_10+c_b1_11) ]
                b_true = level(1).input(level(l).neuron(n).ind_connect(s)).b
            end
            b_hat_final = level(l).neuron(n).syn(s).b_hat_trace(:,:,end)
        end
    end
end

end_t = toc;
tot_t_hrs = end_t/3600

if (run_type == 3)
    lev(2).neu(1).r_on_true = level(2).neuron(1).r_on;
    lev(2).neu(1).r_off_true = level(2).neuron(1).r_off;
    lev(3).neu(1).r_on_true = level(3).neuron(1).r_on;
    lev(3).neu(1).r_off_true = level(3).neuron(1).r_off;
    for s = 1:level(1).N_inputs
        lev(2).neu(1).syn(s).q_i_on_true = level(1).input(s).lambda_on; 
        lev(2).neu(1).syn(s).q_i_off_true = level(1).input(s).lambda_off;
    end
end
for l = 2:N_levels
    lev(l).N_neurons = level(l).N_neurons;
    for n = 1:level(l).N_neurons
        lev(l).neu(n).r_on_init = level(l).neuron(n).r_on_init;
        lev(l).neu(n).r_off_init = level(l).neuron(n).r_off_init;
        for s = 1:level(l).neuron(n).N_connects
            lev(l).neu(n).syn(s).q_i_on_init = level(l).neuron(n).syn(s).q_i_on_init; 
            lev(l).neu(n).syn(s).q_i_off_init = level(l).neuron(n).syn(s).q_i_off_init;
        end
        level(l).neuron(n).y_k_trace = [];
        lev(l).neu(n).crit_reached = level(l).neuron(n).crit_reached; 
        lev(l).neu(n).crit_p = level(l).neuron(n).crit_p;
        lev(l).neu(n).R_on = level(l).neuron(n).r_on_hat;
        lev(l).neu(n).R_off = level(l).neuron(n).r_off_hat;
        lev(l).neu(n).R_on_bar = level(l).neuron(n).ronbar;
        lev(l).neu(n).R_off_bar = level(l).neuron(n).roffbar;
        qonh_array = zeros(1,level(l).neuron(n).N_connects);
        qoffh_array = zeros(1,level(l).neuron(n).N_connects);
        qonb_array = zeros(1,level(l).neuron(n).N_connects);
        qoffb_array = zeros(1,level(l).neuron(n).N_connects);
        for s = 1:level(l).neuron(n).N_connects
            qonh_array(s) = level(l).neuron(n).syn(s).q_i_on_hat;
            qoffh_array(s) = level(l).neuron(n).syn(s).q_i_off_hat;
            qonb_array(s) = level(l).neuron(n).syn(s).qonbar;
            qoffb_array(s) = level(l).neuron(n).syn(s).qoffbar;
        end
        lev(l).neu(n).q_i_on = qonh_array;
        lev(l).neu(n).q_i_off = qoffh_array;
        lev(l).neu(n).q_i_on_bar = qonb_array;
        lev(l).neu(n).q_i_off_bar = qoffb_array;
        lev(l).neu(n).L = level(l).neuron(n).L;
        if (l == 2)
           lev(l).neu(n).L_true = level(l).neuron(n).L_true;
           lev(l).neu(n).L_true_flip = level(l).neuron(n).L_true_flip;
        end
        lev(l).neu(n).x_k_trace = level(l).neuron(n).x_k_trace;
        lev(l).neu(n).x_k_est_trace = level(l).neuron(n).x_k_est_trace;
        lev(l).neu(n).tot_EM_time_mins = level(l).neuron(n).tot_EM_time_mins;
        lev(l).neu(n).tot_EM_steps = level(l).neuron(n).tot_EM_steps;
        lev(l).neu(n).total_input_spikes = level(l).neuron(n).total_input_spikes;
        lev(l).neu(n).total_output_spikes = level(l).neuron(n).total_output_spikes;
    end
end

if (use_cnvg == 1)
    cnvg.r_on_follow = resample(cnvg.r_on_follow,1,100);
    cnvg.r_off_follow = resample(cnvg.r_off_follow,1,100);
    cnvg.q_on_follow = resample(cnvg.q_on_follow',1,100)';
    cnvg.q_off_follow = resample(cnvg.q_off_follow',1,100)';
    
    if (perturb == 3)
        perc_sc(3) = 10*perc_sc(3);
    end
    eval([ 'save ' save_path 'cnvg_percsc_' num2str(perc_sc(perturb)) '_BN_2neuron_s' num2str(seed) '_rs' num2str(rand_seed) '_p' num2str(perturb) '_s' num2str(sample_select) '_r' num2str(rate_select) '.mat cnvg -mat' ])
    eval([ 'save ' save_path 'BN_2neuron_s' num2str(seed) '_rs' num2str(rand_seed) '_p' num2str(perturb) '_s' num2str(sample_select) '_r' num2str(rate_select) '.mat lev tot_t_hrs -mat' ])
else
    eval([ 'save ' save_path 'BN_2neuron_s' num2str(seed) '_rs' num2str(rand_seed) '_p' num2str(perturb) '_s' num2str(sample_select) '_r' num2str(rate_select) '.mat lev tot_t_hrs -mat' ])
end
