function BN_network_small(seed)

% Simulates the FL algorithm in a 3 layer Bayesian Spiking Neuron (BSN) 
% network with 4, 2 and 1 neurons in layers 1, 2 and 3, respectively.

% This function is more streamlined in comparison to 'BN_FL_2Neuron.m' to make the network simulation run faster. This code is commented, but  
% please refer to the comments used in 'BN_FL_2Neuron.m' if parts are unclear as same variable names are reused. 

% inputs: 
% 'seed': seeds the randomisation of the true parameter values for the
% transition and observation rates, as well as the randomisation of the initial parameter estimates. 

% outputs: There are no outputs, however, simulated variables are automatically saved to a .mat file 

% Note rates are given in units of spikes/(0.1 ms). Therefore be aware of the following
% conversion: r = 1 spikes/s = 0.001 spikes/ms = 0.001 spikes/ 10 x 0.1 ms = 0.0001 spikes/(0.1 ms)

% This code includes additional code/checks/options that can be removed for
% efficiency purposes but have been left here. The preference has been to
% just give adequate comments. Follow the comments in a sequential order to
% make the most sense of the code. It is recommended to make your own simplified code
% but to use this as a guide.

% For help email: levink@unimelb.edu.au

% Copyright 15/11/2013 Levin Kuhlmann

tic;

dbstop if error

dt = 0.1; % time step in ms
T = 1000000; % number of time steps
T_last = 900000; % time step after which output spikes are calculated
true_init = 0; % 0 means parameter estimates are initialised randomly in the same uniform range as the true parameter values, 
% 1 means the initial parameter estimates are set to the true values.
eta = 1/100000; % forgetting factor
T_stat = 1*10^5; % the number of time steps before which the parameter estimates are allowed to be updated.
N_track = 100000; % the number of time steps over which the optional figure tracks the simulated variables.
p_up_sc = 0.75; % theta_U from the paper
p_down_sc = 0.25; % theta_D from the paper
check_test = 1; % 1: means hierarchical causal models are simulated as in the paper. 0: means a noisy bar scenario is simulated as in Deneve 2008b, however,
% I don't know if I interpretted her scenario properly and FL did not do
% well with my interpretation so it was just left out of the paper
rand('state',seed)

%save_path = 'C:/my_files/spikeSim_research/bayesian_neuron/on_big_data_1mil/'
save_path = 'on_big_data_1mil/'
N_levels = 3;
% set number of neurons in each layer
level(1).N_neurons = 4;
level(2).N_neurons = 2;
level(3).N_neurons = 1;
% set spiking thresholds
level(1).neuron(1).g_o = 1.45;
level(1).neuron(2).g_o = 1.45;
level(1).neuron(3).g_o = 1.45;
level(1).neuron(4).g_o = 1.45;
level(2).neuron(1).g_o = 1.45;
level(2).neuron(2).g_o = 1.45;
level(3).neuron(1).g_o = 1.45;
if (check_test == 0) % specify the sizes of the bar array
    N_bar_posnx = 4;
    N_bar_posny = 3;
    N_bar_posn = 3;
    bar_length = 4;
    N_bar_ld = 2;
elseif (check_test == 1) % here the specified sizes of the bar array essentially just mean 20 (2*2*5) independent inputs 
    % were used to drive each neuron in the first layer (note first layer
    % in this code refers to the first layer of neurons not the input
    % layer)
    N_bar_posnx = 5;
    N_bar_posny = 2;
    N_bar_posn = 2;
    bar_length = 5;
    N_bar_ld = 2;    
end
level(1).neuron(1).N_inputs = N_bar_ld*N_bar_posn*bar_length;
level(1).neuron(2).N_inputs = N_bar_ld*N_bar_posn*bar_length;
level(1).neuron(3).N_inputs = N_bar_ld*N_bar_posn*bar_length;
level(1).neuron(4).N_inputs = N_bar_ld*N_bar_posn*bar_length;
level(2).neuron(1).N_inputs = 2;
level(2).neuron(2).N_inputs = 2;
level(3).neuron(1).N_inputs = 2;

% true parameters
if (check_test == 0) % set up true rates as specified in Deneve 2008b
    level(1).neuron(1).q_on_t = zeros(N_bar_posn,bar_length,N_bar_ld);
    level(1).neuron(2).q_on_t = zeros(N_bar_posn,bar_length,N_bar_ld);
    level(1).neuron(3).q_on_t = zeros(N_bar_posn,bar_length,N_bar_ld);
    level(1).neuron(4).q_on_t = zeros(N_bar_posn,bar_length,N_bar_ld);
    level(1).neuron(1).q_on_t(:,:,1) = 0.08*ones(N_bar_posn,bar_length,1);
    level(1).neuron(2).q_on_t(:,:,1) = 0.08*ones(N_bar_posn,bar_length,1);
    level(1).neuron(3).q_on_t(:,:,1) = 0.08*ones(N_bar_posn,bar_length,1);
    level(1).neuron(4).q_on_t(:,:,1) = 0.08*ones(N_bar_posn,bar_length,1);
    level(1).neuron(1).q_on_t(:,:,2) = 0.04*ones(N_bar_posn,bar_length,1);
    level(1).neuron(2).q_on_t(:,:,2) = 0.04*ones(N_bar_posn,bar_length,1);
    level(1).neuron(3).q_on_t(:,:,2) = 0.04*ones(N_bar_posn,bar_length,1);
    level(1).neuron(4).q_on_t(:,:,2) = 0.04*ones(N_bar_posn,bar_length,1);

    % don't know true values for higher level neurons
    level(2).neuron(1).q_on_t = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(2).neuron(2).q_on_t = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(3).neuron(1).q_on_t = (0.001+rand(1,level(2).N_neurons)*0.999);

    level(1).neuron(1).q_off_t = zeros(N_bar_posn,bar_length,N_bar_ld);
    level(1).neuron(2).q_off_t = zeros(N_bar_posn,bar_length,N_bar_ld);
    level(1).neuron(3).q_off_t = zeros(N_bar_posn,bar_length,N_bar_ld);
    level(1).neuron(4).q_off_t = zeros(N_bar_posn,bar_length,N_bar_ld);
    level(1).neuron(1).q_off_t(:,:,1) = 0.02*ones(N_bar_posn,bar_length,1);
    level(1).neuron(2).q_off_t(:,:,1) = 0.02*ones(N_bar_posn,bar_length,1);
    level(1).neuron(3).q_off_t(:,:,1) = 0.02*ones(N_bar_posn,bar_length,1);
    level(1).neuron(4).q_off_t(:,:,1) = 0.02*ones(N_bar_posn,bar_length,1);
    level(1).neuron(1).q_off_t(:,:,2) = 0.075*ones(N_bar_posn,bar_length,1);
    level(1).neuron(2).q_off_t(:,:,2) = 0.075*ones(N_bar_posn,bar_length,1);
    level(1).neuron(3).q_off_t(:,:,2) = 0.075*ones(N_bar_posn,bar_length,1);
    level(1).neuron(4).q_off_t(:,:,2) = 0.075*ones(N_bar_posn,bar_length,1);

    % don't know true values for higher level neurons
    level(2).neuron(1).q_off_t = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(2).neuron(2).q_off_t = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(3).neuron(1).q_off_t = (0.001+rand(1,level(2).N_neurons)*0.999);

    level(1).neuron(1).r_on_t = 0.05;
    level(1).neuron(2).r_on_t = 0.05;
    level(1).neuron(3).r_on_t = 0.05;
    level(1).neuron(4).r_on_t = 0.05;
    level(2).neuron(1).r_on_t = 0.025;
    level(2).neuron(2).r_on_t = 0.025;
    level(3).neuron(1).r_on_t = 0.005;
    level(1).neuron(1).r_off_t = 0.1;
    level(1).neuron(2).r_off_t = 0.1;
    level(1).neuron(3).r_off_t = 0.1;
    level(1).neuron(4).r_off_t = 0.1;
    level(2).neuron(1).r_off_t = 0.05;
    level(2).neuron(2).r_off_t = 0.05;
    level(3).neuron(1).r_off_t = 0.01;
elseif (check_test == 1) % set up true rates selected uniformly within the physiologically plausible ranges given in the paper.
    level(1).neuron(1).q_on_t = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(2).q_on_t = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(3).q_on_t = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(4).q_on_t = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(2).neuron(1).q_on_t = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(2).neuron(2).q_on_t = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(3).neuron(1).q_on_t = (0.001+rand(1,level(2).N_neurons)*0.999);
    level(1).neuron(1).q_off_t = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(2).q_off_t = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(3).q_off_t = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(4).q_off_t = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(2).neuron(1).q_off_t = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(2).neuron(2).q_off_t = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(3).neuron(1).q_off_t = (0.001+rand(1,level(2).N_neurons)*0.999);
    level(1).neuron(1).r_on_t = (0.001+rand(1)*0.114);
    level(1).neuron(2).r_on_t = level(1).neuron(1).r_on_t;
    level(1).neuron(3).r_on_t = level(1).neuron(1).r_on_t;
    level(1).neuron(4).r_on_t = level(1).neuron(1).r_on_t;
    level(2).neuron(1).r_on_t = level(1).neuron(1).r_on_t;
    level(2).neuron(2).r_on_t = level(1).neuron(1).r_on_t;
    level(3).neuron(1).r_on_t = level(1).neuron(1).r_on_t;
    level(1).neuron(1).r_off_t = (0.001+rand(1)*0.114);
    level(1).neuron(2).r_off_t = level(1).neuron(1).r_off_t;
    level(1).neuron(3).r_off_t = level(1).neuron(1).r_off_t;
    level(1).neuron(4).r_off_t = level(1).neuron(1).r_off_t;
    level(2).neuron(1).r_off_t = level(1).neuron(1).r_off_t;
    level(2).neuron(2).r_off_t = level(1).neuron(1).r_off_t;
    level(3).neuron(1).r_off_t = level(1).neuron(1).r_off_t;
end

% parameter estimates
if (true_init == 1) % set initial estimates equal to true values
    for l = 1:N_levels
        for n = 1:level(l).N_neurons
            level(l).neuron(n).q_on = level(l).neuron(n).q_on_t;
            level(l).neuron(n).q_off = level(l).neuron(n).q_off_t;
            level(l).neuron(n).r_on = level(l).neuron(n).r_on_t;
            level(l).neuron(n).r_off = level(l).neuron(n).r_off_t;
        end
    end
elseif (true_init == 0) % initialise rate estimates selected uniformly within the physiologically plausible ranges used for the true parameters.
    level(1).neuron(1).q_on = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(2).q_on = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(3).q_on = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(4).q_on = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(2).neuron(1).q_on = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(2).neuron(2).q_on = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(3).neuron(1).q_on = (0.001+rand(1,level(2).N_neurons)*0.999);
    level(1).neuron(1).q_off = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(2).q_off = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(3).q_off = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(1).neuron(4).q_off = (0.001+rand(N_bar_posn,bar_length,N_bar_ld)*0.999);
    level(2).neuron(1).q_off = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(2).neuron(2).q_off = (0.001+rand(1,level(1).N_neurons/2)*0.999);
    level(3).neuron(1).q_off = (0.001+rand(1,level(2).N_neurons)*0.999);
    level(1).neuron(1).r_on = (0.001+rand(1)*0.114);
    level(1).neuron(2).r_on = (0.001+rand(1)*0.114);
    level(1).neuron(3).r_on = (0.001+rand(1)*0.114);
    level(1).neuron(4).r_on = (0.001+rand(1)*0.114);
    level(2).neuron(1).r_on = (0.001+rand(1)*0.114);
    level(2).neuron(2).r_on = (0.001+rand(1)*0.114);
    level(3).neuron(1).r_on = (0.001+rand(1)*0.114);
    level(1).neuron(1).r_off = (0.001+rand(1)*0.114);
    level(1).neuron(2).r_off = (0.001+rand(1)*0.114);
    level(1).neuron(3).r_off = (0.001+rand(1)*0.114);
    level(1).neuron(4).r_off = (0.001+rand(1)*0.114);
    level(2).neuron(1).r_off = (0.001+rand(1)*0.114);
    level(2).neuron(2).r_off = (0.001+rand(1)*0.114);
    level(3).neuron(1).r_off = (0.001+rand(1)*0.114);
end
for l = 1:N_levels
    for n = 1:level(l).N_neurons
        level(l).neuron(n).input = zeros(size(level(l).neuron(n).q_on));
    end
end

% define a and b matrices for making the input
a_bar = [ (1-level(1).neuron(1).r_on_t*dt) level(1).neuron(1).r_on_t*dt ; level(1).neuron(1).r_off_t*dt (1-level(1).neuron(1).r_off_t*dt) ];
a_line = [ (1-level(2).neuron(1).r_on_t*dt) level(2).neuron(1).r_on_t*dt ; level(2).neuron(1).r_off_t*dt (1-level(2).neuron(1).r_off_t*dt) ];
a_cont = [ (1-level(3).neuron(1).r_on_t*dt) level(3).neuron(1).r_on_t*dt ; level(3).neuron(1).r_off_t*dt (1-level(3).neuron(1).r_off_t*dt) ];
if (check_test == 0)
    b_ON = [ (1-level(1).neuron(1).q_off_t(1,1,1)*dt) level(1).neuron(1).q_off_t(1,1,1)*dt ; (1-level(1).neuron(1).q_on_t(1,1,1)*dt) level(1).neuron(1).q_on_t(1,1,1)*dt ];
    b_OFF = [ (1-level(1).neuron(1).q_off_t(1,1,2)*dt) level(1).neuron(1).q_off_t(1,1,2)*dt ; (1-level(1).neuron(1).q_on_t(1,1,2)*dt) level(1).neuron(1).q_on_t(1,1,2)*dt ];
elseif (check_test == 1)  
    for n = 1:level(1).N_neurons
        for s = 1:level(1).neuron(n).N_inputs
            level(1).neuron(n).syn(s).b = [ (1-level(1).neuron(n).q_off_t(s)*dt) level(1).neuron(n).q_off_t(s)*dt ; (1-level(1).neuron(n).q_on_t(s)*dt) level(1).neuron(n).q_on_t(s)*dt ];
        end
    end   
end

% trace arrays for each neuron and initialisation
for l = 1:N_levels
    for n = 1:level(l).N_neurons
        level(l).neuron(n).x_k_trace = zeros(1,N_track); % neuron hidden state
        level(l).neuron(n).x_k_est_trace = zeros(1,N_track);
        level(l).neuron(n).y_k_trace = 0;
        level(l).neuron(n).output_spike_count = 0;
        level(l).neuron(n).L = zeros(1,N_track);
        level(l).neuron(n).p_on_trace = 0.5*ones(1,N_track);
        level(l).neuron(n).L(1) = log(level(l).neuron(n).r_on/level(l).neuron(n).r_off);
        level(l).neuron(n).G = level(l).neuron(n).L(1);
        if (l == 1)
            level(l).neuron(n).L_true = level(l).neuron(n).L;
            level(l).neuron(n).L_true_flip = level(l).neuron(n).L;
        end
        level(l).neuron(n).t_on = 0.5;
        level(l).neuron(n).N_on_off = 0;
        level(l).neuron(n).N_off_on = 0;
        level(l).neuron(n).N_on = zeros(size(level(l).neuron(n).input));
        level(l).neuron(n).N = zeros(size(level(l).neuron(n).input));
        level(l).neuron(n).x_c = 0;
    end
end
input.stim = zeros(N_bar_posny,N_bar_posnx); % used to store the current input stimulus to the network in terms of bar positions
input.stim_full = zeros(N_bar_posny,N_bar_posnx*bar_length); % used to store the current input stimulus to the network in terms of input pixels
X_bar = zeros(N_bar_posny,N_bar_posnx);
X_line = zeros(1,2);
X_cont = 0;
if (check_test == 1)
    X = 0;
end
for t = 2:T
    t_mod = mod(t-1,N_track) + 1;
    t_mod_m1 = mod(t-2,N_track) + 1;
    t_mod_m10000 = mod(t-5000,N_track) + 1;
    
    if (t_mod_m10000 < t_mod_m1)
        p_ind10000 = [ t_mod_m10000:t_mod_m1 ];
    else
        p_ind10000 = [ 1:t_mod_m1 t_mod_m10000:N_track ];
    end
     
    % evolve HMM hidden states
    if (check_test == 0) % noisy bar simulation of Deneve 2008b
        ind_X0 = find(X_bar(:) == 0);
        ind_X1 = find(X_bar(:) == 1);
        if ~isempty(ind_X0)
            X_bar(ind_X0) = (rand(1,numel(ind_X0)) > a_bar(1,1));       
        end
        if ~isempty(ind_X1)
            X_bar(ind_X1) = (rand(1,numel(ind_X1)) > a_bar(2,1));       
        end
        ind_X0 = find(X_line(:) == 0);
        ind_X1 = find(X_line(:) == 1);
        if ~isempty(ind_X0)
            X_line(ind_X0) = (rand(1,numel(ind_X0)) > a_line(1,1));       
        end
        if ~isempty(ind_X1)
            X_line(ind_X1) = (rand(1,numel(ind_X1)) > a_line(2,1));       
        end
        if (X_cont == 0)
            X_cont = (rand(1) > a_cont(1,1));
        else
            X_cont = (rand(1) > a_cont(2,1));
        end
        X_bar_neurons = min(sum(X_bar,1),1);
    elseif (check_test == 1) % hierarchical causal model
        if (X == 0)
            X = (rand(1) > a_bar(1,1));
        else
            X = (rand(1) > a_bar(2,1));
        end
        % all hidden states equal since using hierarchical causal model
        X_bar_neurons = X*ones(1,level(1).N_neurons);
        X_bar = X*ones(size(X_bar));
        X_line = X*ones(1,level(2).N_neurons);
        X_cont = X;        
    end
    for nn = 1:level(1).N_neurons
        level(1).neuron(nn).x_k_trace(t_mod) = X_bar_neurons(nn);
    end
    for nn = 1:level(2).N_neurons
        level(2).neuron(nn).x_k_trace(t_mod) = X_line(nn);
    end
    level(3).neuron(1).x_k_trace(t_mod) = X_cont;
    % compute input stimulus
    if (check_test == 0) % for noisy bars case
        input.stim = X_bar;
        if (X_line(1) == 1)
            input.stim(2,1:2) = ones(1,2);
        end
        if (X_line(2) == 1)
            input.stim(2,3:4) = ones(1,2);
        end
        if (X_cont == 1)
            input.stim(2,1:4) = ones(1,4);
        end
        for j = 1:N_bar_posnx*bar_length
            input.stim_full(:,j) = input.stim(:,ceil(j/bar_length));
        end
        % work out input spikes to the first layer
        for nn = 1:level(1).N_neurons
            ind_j = [1:bar_length]+(nn-1)*bar_length;
            local_input = input.stim_full(:,ind_j);
            tmp_ON = zeros(size(local_input));
            tmp_OFF = zeros(size(local_input));
            ind_I0 = find(local_input(:) == 0);
            ind_I1 = find(local_input(:) == 1);
            if ~isempty(ind_I0)
                tmp_ON(ind_I0) = ( rand(1,numel(ind_I0)) > b_ON(1,1) );
            end
            if ~isempty(ind_I1)
                tmp_ON(ind_I1) = ( rand(1,numel(ind_I1)) > b_ON(2,1) );
            end
            if ~isempty(ind_I0)
                tmp_OFF(ind_I0) = ( rand(1,numel(ind_I0)) > b_OFF(1,1) );
            end
            if ~isempty(ind_I1)
                tmp_OFF(ind_I1) = ( rand(1,numel(ind_I1)) > b_OFF(2,1) );
            end
            level(1).neuron(nn).input(:,:,1) = tmp_ON;
            level(1).neuron(nn).input(:,:,2) = tmp_OFF;
        end
    elseif (check_test == 1) % for hierarchical causal case. here all inputs are independent.
        for n = 1:level(1).N_neurons
            for s = 1:level(1).neuron(n).N_inputs
                if (X == 0)
                    level(1).neuron(n).input(s) = ( rand(1) > level(1).neuron(n).syn(s).b(1,1) );
                elseif (X == 1)
                    level(1).neuron(n).input(s) = ( rand(1) > level(1).neuron(n).syn(s).b(2,1) );
                end
            end
        end
    end
    % update neurons
    for l = 1:N_levels
        for n = 1:level(l).N_neurons
            p_mx = max(level(l).neuron(n).p_on_trace(p_ind10000));
            p_mn = min(level(l).neuron(n).p_on_trace(p_ind10000));
            p_up = p_mn + p_up_sc*(p_mx-p_mn);
            p_down = p_mn + p_down_sc*(p_mx-p_mn);
            p_on = exp(level(l).neuron(n).L(t_mod_m1))/(1+exp(level(l).neuron(n).L(t_mod_m1)));
            level(l).neuron(n).p_on_trace(t_mod) = p_on;
            x_pr = level(l).neuron(n).x_c;
            level(l).neuron(n).N_on_off = (1-eta)*level(l).neuron(n).N_on_off;
            level(l).neuron(n).N_off_on = (1-eta)*level(l).neuron(n).N_off_on;
            if (p_on > p_up)
                level(l).neuron(n).x_c = 1;
                if (x_pr == 0)
                    level(l).neuron(n).N_off_on = eta + level(l).neuron(n).N_off_on;
                end
            elseif (p_on < p_down)
                level(l).neuron(n).x_c = 0;
                if (x_pr == 1)
                    level(l).neuron(n).N_on_off = eta + level(l).neuron(n).N_on_off;
                end
            end
            level(l).neuron(n).x_k_est_trace(t_mod) = level(l).neuron(n).x_c;
            level(l).neuron(n).t_on = eta*(level(l).neuron(n).x_c + 10^(-15)) + (1-eta)*level(l).neuron(n).t_on;
            level(l).neuron(n).N_on = eta*level(l).neuron(n).input*level(l).neuron(n).x_c + (1-eta)*level(l).neuron(n).N_on;
            level(l).neuron(n).N = eta*level(l).neuron(n).input + (1-eta)*level(l).neuron(n).N;
            
            if (t > T_stat) % keep parameters fixed for first T_stat time steps for phi to stabilise
                level(l).neuron(n).r_on = max(level(l).neuron(n).N_off_on/(dt*(1 - level(l).neuron(n).t_on)),0.0001);
                level(l).neuron(n).r_off = max(level(l).neuron(n).N_on_off/(dt*level(l).neuron(n).t_on),0.0001);
                level(l).neuron(n).q_on = level(l).neuron(n).N_on/(dt*level(l).neuron(n).t_on) + 10^(-15);
                level(l).neuron(n).q_off = (level(l).neuron(n).N - level(l).neuron(n).N_on)/(dt*(1 - level(l).neuron(n).t_on)) + 10^(-15);
            end
            
            % update bayesian neuron
            dL_spikes = sum( level(l).neuron(n).input(:).*log(level(l).neuron(n).q_on(:)./level(l).neuron(n).q_off(:)) + (1-level(l).neuron(n).input(:)).*log((1-level(l).neuron(n).q_on(:)*dt)./(1-level(l).neuron(n).q_off(:)*dt)) );
            level(l).neuron(n).L(t_mod) = level(l).neuron(n).L(t_mod_m1) + log(1+dt*(-level(l).neuron(n).r_off+level(l).neuron(n).r_on*exp(-level(l).neuron(n).L(t_mod_m1)))) - log(1+dt*(-level(l).neuron(n).r_on+level(l).neuron(n).r_off*exp(level(l).neuron(n).L(t_mod_m1)))) + dL_spikes;
            if (l == 1)
                dL_true_spikes = sum( level(l).neuron(n).input(:).*log(level(l).neuron(n).q_on_t(:)./level(l).neuron(n).q_off_t(:)) + (1-level(l).neuron(n).input(:)).*log((1-level(l).neuron(n).q_on_t(:)*dt)./(1-level(l).neuron(n).q_off_t(:)*dt)) );
                level(l).neuron(n).L_true(t_mod) = level(l).neuron(n).L_true(t_mod_m1) + log(1+dt*(-level(l).neuron(n).r_off_t+level(l).neuron(n).r_on_t*exp(-level(l).neuron(n).L_true(t_mod_m1)))) - log(1+dt*(-level(l).neuron(n).r_on_t+level(l).neuron(n).r_off_t*exp(level(l).neuron(n).L_true(t_mod_m1)))) + dL_true_spikes;
                dL_true_flip_spikes = sum( level(l).neuron(n).input(:).*log(level(l).neuron(n).q_off_t(:)./level(l).neuron(n).q_on_t(:)) + (1-level(l).neuron(n).input(:)).*log((1-level(l).neuron(n).q_off_t(:)*dt)./(1-level(l).neuron(n).q_on_t(:)*dt)) );
                level(l).neuron(n).L_true_flip(t_mod) = level(l).neuron(n).L_true_flip(t_mod_m1) + log(1+dt*(-level(l).neuron(n).r_on_t+level(l).neuron(n).r_off_t*exp(-level(l).neuron(n).L_true_flip(t_mod_m1)))) - log(1+dt*(-level(l).neuron(n).r_off_t+level(l).neuron(n).r_on_t*exp(level(l).neuron(n).L_true_flip(t_mod_m1)))) + dL_true_flip_spikes;
            end
            level(l).neuron(n).G = level(l).neuron(n).G + log(1+dt*(-level(l).neuron(n).r_off+level(l).neuron(n).r_on*exp(-level(l).neuron(n).G))) - log(1+dt*(-level(l).neuron(n).r_on+level(l).neuron(n).r_off*exp(level(l).neuron(n).G))) + level(l).neuron(n).g_o*level(l).neuron(n).y_k_trace;
            if (level(l).neuron(n).L(t_mod) > level(l).neuron(n).G + level(l).neuron(n).g_o/2)
                level(l).neuron(n).y_k_trace = 1;
                if (t > T_last)
                    level(l).neuron(n).output_spike_count = level(l).neuron(n).output_spike_count + 1;
                end
            else
                level(l).neuron(n).y_k_trace = 0;
            end
            
        end % end n loop
        
        % rearrange inputs to outputs
        if (l == 1)
            level(l+1).neuron(1).input = [ level(l).neuron(1).y_k_trace level(l).neuron(2).y_k_trace ];
            level(l+1).neuron(2).input = [ level(l).neuron(3).y_k_trace level(l).neuron(4).y_k_trace ];
        elseif (l == 2)
            level(l+1).neuron(1).input = [ level(l).neuron(1).y_k_trace level(l).neuron(2).y_k_trace ];
        end
        
    end % end l loop  
 end % end t loop

% store key variables in a smaller struct for saving. 
for l = 1:N_levels
    lev(l).N_neurons = level(l).N_neurons;
    for n = 1:level(l).N_neurons 
        lev(l).neu(n).r_on_t = level(l).neuron(n).r_on_t;
        lev(l).neu(n).r_off_t = level(l).neuron(n).r_off_t;
        lev(l).neu(n).q_on_t = level(l).neuron(n).q_on_t;
        lev(l).neu(n).q_off_t = level(l).neuron(n).q_off_t;
        lev(l).neu(n).r_on = level(l).neuron(n).r_on;
        lev(l).neu(n).r_off = level(l).neuron(n).r_off;
        lev(l).neu(n).q_on = level(l).neuron(n).q_on;
        lev(l).neu(n).q_off = level(l).neuron(n).q_off;
        lev(l).neu(n).L = level(l).neuron(n).L;
        if (l == 1)
           lev(l).neu(n).L_true = level(l).neuron(n).L_true;
           lev(l).neu(n).L_true_flip = level(l).neuron(n).L_true_flip;
        end
        lev(l).neu(n).x_k_trace = level(l).neuron(n).x_k_trace;
        lev(l).neu(n).x_k_est_trace = level(l).neuron(n).x_k_est_trace;
        lev(l).neu(n).output_spike_count = level(l).neuron(n).output_spike_count; 
    end
end

total_time_hrs = toc/3600

% save key variables
eval([ 'save ' save_path 'BN_network_s' num2str(seed) '.mat lev total_time_hrs -mat' ])
